@@ -67,14 +67,24 @@
 #include <string.h>
 
 #include <sys/shm.h>
-#include <sys/sem.h>
+
 #include <corosync/corotypes.h>
 #include <corosync/list.h>
 
 #include <corosync/coroipc_types.h>
+#include <corosync/hdb.h>
 #include <corosync/coroipcs.h>
 #include <corosync/coroipc_ipc.h>
 
+#define LOGSYS_UTILS_ONLY 1
+#include <corosync/engine/logsys.h>
+
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+#include <semaphore.h>
+#else
+#include <sys/sem.h>
+#endif
+
 #ifndef MSG_NOSIGNAL
 #define MSG_NOSIGNAL 0
 #endif
@@ -84,7 +94,7 @@
 #define MSG_SEND_LOCKED		0
 #define MSG_SEND_UNLOCKED	1
 
-static struct coroipcs_init_state *api;
+static struct coroipcs_init_state_v2 *api = NULL;
 
 DECLARE_LIST_INIT (conn_info_list_head);
 
@@ -100,6 +110,7 @@ struct zcb_mapped {
 	size_t size;
 };
 
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 #if defined(_SEM_SEMUN_UNDEFINED)
 union semun {
 	int val;
@@ -108,6 +119,8 @@ union semun {
 	struct seminfo *__buf;
 };
 #endif
+#endif
+
 
 enum conn_state {
 	CONN_STATE_THREAD_INACTIVE = 0,
@@ -121,14 +134,18 @@ enum conn_state {
 struct conn_info {
 	int fd;
 	pthread_t thread;
+	pid_t client_pid;
 	pthread_attr_t thread_attr;
 	unsigned int service;
 	enum conn_state state;
 	int notify_flow_control_enabled;
+	int flow_control_state;
 	int refcount;
-	key_t shmkey;
+	hdb_handle_t stats_handle;
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	key_t semkey;
 	int semid;
+#endif
 	unsigned int pending_semops;
 	pthread_mutex_t mutex;
 	struct control_buffer *control_buffer;
@@ -159,6 +176,76 @@ static void ipc_disconnect (struct conn_info *conn_info);
 static void msg_send (void *conn, const struct iovec *iov, unsigned int iov_len,
 		      int locked);
 
+static void _corosync_ipc_init(void);
+
+#define log_printf(level, format, args...) \
+do { \
+	if (api->log_printf) \
+        api->log_printf ( \
+                LOGSYS_ENCODE_RECID(level, \
+                                    api->log_subsys_id, \
+                                    LOGSYS_RECID_LOG), \
+                __FUNCTION__, __FILE__, __LINE__, \
+                (const char *)format, ##args); \
+	else \
+        api->old_log_printf ((const char *)format, ##args); \
+} while (0)
+
+static hdb_handle_t dummy_stats_create_connection (
+	const char *name,
+	pid_t pid,
+	int fd)
+{
+	return (0ULL);
+}
+
+static void dummy_stats_destroy_connection (
+	hdb_handle_t handle)
+{
+}
+
+static void dummy_stats_update_value (
+	hdb_handle_t handle,
+	const char *name,
+	const void *value,
+	size_t value_size)
+{
+}
+
+static void dummy_stats_increment_value (
+	hdb_handle_t handle,
+	const char *name)
+{
+}
+
+static void sem_post_exit_thread (struct conn_info *conn_info)
+{
+#if _POSIX_THREAD_PROCESS_SHARED < 1
+	struct sembuf sop;
+#endif
+	int res;
+
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+retry_semop:
+	res = sem_post (&conn_info->control_buffer->sem0);
+	if (res == -1 && errno == EINTR) {
+		api->stats_increment_value (conn_info->stats_handle, "sem_retry_count");
+		goto retry_semop;
+	}
+#else
+	sop.sem_num = 0;
+	sop.sem_op = 1;
+	sop.sem_flg = 0;
+
+retry_semop:
+	res = semop (conn_info->semid, &sop, 1);
+	if ((res == -1) && (errno == EINTR || errno == EAGAIN)) {
+		api->stats_increment_value (conn_info->stats_handle, "sem_retry_count");
+		goto retry_semop;
+	}
+#endif
+}
+
 static int
 memory_map (
 	const char *path,
@@ -327,6 +414,7 @@ static inline int zcb_alloc (
 		size,
 		addr);
 	if (res == -1) {
+		free (zcb_mapped);
 		return (-1);
 	}
 
@@ -392,7 +480,7 @@ static inline int conn_info_destroy (struct conn_info *conn_info)
 	}
 
 	if (conn_info->state == CONN_STATE_THREAD_ACTIVE) {
-		pthread_kill (conn_info->thread, SIGUSR1);
+		sem_post_exit_thread (conn_info);
 		return (0);
 	}
 
@@ -401,6 +489,7 @@ static inline int conn_info_destroy (struct conn_info *conn_info)
 	 * Retry library exit function if busy
 	 */
 	if (conn_info->state == CONN_STATE_THREAD_DESTROYED) {
+		api->stats_destroy_connection (conn_info->stats_handle);
 		res = api->exit_fn_get (conn_info->service) (conn_info);
 		if (res == -1) {
 			api->serialize_unlock ();
@@ -419,13 +508,19 @@ static inline int conn_info_destroy (struct conn_info *conn_info)
 	list_del (&conn_info->list);
 	pthread_mutex_unlock (&conn_info->mutex);
 
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+	sem_destroy (&conn_info->control_buffer->sem0);
+	sem_destroy (&conn_info->control_buffer->sem1);
+	sem_destroy (&conn_info->control_buffer->sem2);
+#else
+	semctl (conn_info->semid, 0, IPC_RMID);
+#endif
 	/*
 	 * Destroy shared memory segment and semaphore
 	 */
 	res = munmap ((void *)conn_info->control_buffer, conn_info->control_size);
 	res = munmap ((void *)conn_info->request_buffer, conn_info->request_size);
 	res = munmap ((void *)conn_info->response_buffer, conn_info->response_size);
-	semctl (conn_info->semid, 0, IPC_RMID);
 
 	/*
 	 * Free allocated data needed to retry exiting library IPC connection
@@ -521,7 +616,9 @@ static inline void zerocopy_operations_process (
 static void *pthread_ipc_consumer (void *conn)
 {
 	struct conn_info *conn_info = (struct conn_info *)conn;
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	struct sembuf sop;
+#endif
 	int res;
 	coroipc_request_header_t *header;
 	coroipc_response_header_t coroipc_response_header;
@@ -536,22 +633,37 @@ static void *pthread_ipc_consumer (void *conn)
 #endif
 
 	for (;;) {
-		sop.sem_num = 0;
-		sop.sem_op = -1;
-		sop.sem_flg = 0;
-retry_semop:
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+retry_semwait:
+		res = sem_wait (&conn_info->control_buffer->sem0);
 		if (ipc_thread_active (conn_info) == 0) {
 			coroipcs_refcount_dec (conn_info);
 			pthread_exit (0);
 		}
+		if ((res == -1) && (errno == EINTR)) {
+			api->stats_increment_value (conn_info->stats_handle, "sem_retry_count");
+			goto retry_semwait;
+		}
+#else
+
+		sop.sem_num = 0;
+		sop.sem_op = -1;
+		sop.sem_flg = 0;
+retry_semop:
 		res = semop (conn_info->semid, &sop, 1);
+		if (ipc_thread_active (conn_info) == 0) {
+			coroipcs_refcount_dec (conn_info);
+			pthread_exit (0);
+		}
 		if ((res == -1) && (errno == EINTR || errno == EAGAIN)) {
+			api->stats_increment_value (conn_info->stats_handle, "sem_retry_count");
 			goto retry_semop;
 		} else
 		if ((res == -1) && (errno == EINVAL || errno == EIDRM)) {
 			coroipcs_refcount_dec (conn_info);
 			pthread_exit (0);
 		}
+#endif
 
 		zerocopy_operations_process (conn_info, &header, &new_message);
 		/*
@@ -568,14 +680,28 @@ retry_semop:
 			header,
 			conn_info->sending_allowed_private_data);
 
+		/*
+		 * This happens when the message contains some kind of invalid
+		 * parameter, such as an invalid size
+		 */
+		if (send_ok == -1) {
+			coroipc_response_header.size = sizeof (coroipc_response_header_t);
+			coroipc_response_header.id = 0;
+			coroipc_response_header.error = CS_ERR_INVALID_PARAM;
+			coroipcs_response_send (conn_info,
+				&coroipc_response_header,
+				sizeof (coroipc_response_header_t));
+		} else 
 		if (send_ok) {
 			api->serialize_lock();
+			api->stats_increment_value (conn_info->stats_handle, "requests");
 			api->handler_fn_get (conn_info->service, header->id) (conn_info, header);
 			api->serialize_unlock();
 		} else {
 			/*
 			 * Overload, tell library to retry
 			 */
+			api->stats_increment_value (conn_info->stats_handle, "sem_retry_count");
 			coroipc_response_header.size = sizeof (coroipc_response_header_t);
 			coroipc_response_header.id = 0;
 			coroipc_response_header.error = CS_ERR_TRY_AGAIN;
@@ -603,9 +729,11 @@ req_setup_send (
 retry_send:
 	res = send (conn_info->fd, &res_setup, sizeof (mar_res_setup_t), MSG_WAITALL);
 	if (res == -1 && errno == EINTR) {
+		api->stats_increment_value (conn_info->stats_handle, "send_retry_count");
 		goto retry_send;
 	} else
 	if (res == -1 && errno == EAGAIN) {
+		api->stats_increment_value (conn_info->stats_handle, "send_retry_count");
 		goto retry_send;
 	}
 	return (0);
@@ -650,6 +778,7 @@ req_setup_recv (
 retry_recv:
 	res = recvmsg (conn_info->fd, &msg_recv, MSG_NOSIGNAL);
 	if (res == -1 && errno == EINTR) {
+		api->stats_increment_value (conn_info->stats_handle, "recv_retry_count");
 		goto retry_recv;
 	} else
 	if (res == -1 && errno != EAGAIN) {
@@ -684,6 +813,7 @@ retry_recv:
 		if (getpeerucred (conn_info->fd, &uc) == 0) {
 			euid = ucred_geteuid (uc);
 			egid = ucred_getegid (uc);
+			conn_info->client_pid = ucred_getpid (uc);
 			if (api->security_valid (euid, egid)) {
 				authenticated = 1;
 			}
@@ -699,6 +829,10 @@ retry_recv:
 		uid_t euid;
 		gid_t egid;
 
+		/*
+		 * TODO get the peer's pid.
+		 * conn_info->client_pid = ?;
+		 */
 		euid = -1;
 		egid = -1;
 		if (getpeereid (conn_info->fd, &euid, &egid) == 0) {
@@ -716,6 +850,7 @@ retry_recv:
 	assert (cmsg);
 	cred = (struct ucred *)CMSG_DATA (cmsg);
 	if (cred) {
+		conn_info->client_pid = cred->pid;
 		if (api->security_valid (cred->uid, cred->gid)) {
 			authenticated = 1;
 		}
@@ -723,11 +858,11 @@ retry_recv:
 
 #else /* no credentials */
 	authenticated = 1;
- 	api->log_printf ("Platform does not support IPC authentication.  Using no authentication\n");
+ 	log_printf (LOGSYS_LEVEL_ERROR, "Platform does not support IPC authentication.  Using no authentication\n");
 #endif /* no credentials */
 
 	if (authenticated == 0) {
-		api->log_printf ("Invalid IPC credentials.\n");
+		log_printf (LOGSYS_LEVEL_ERROR, "Invalid IPC credentials.\n");
 		ipc_disconnect (conn_info);
 		return (-1);
  	}
@@ -755,7 +890,7 @@ static void ipc_disconnect (struct conn_info *conn_info)
 	conn_info->state = CONN_STATE_THREAD_REQUEST_EXIT;
 	pthread_mutex_unlock (&conn_info->mutex);
 
-	pthread_kill (conn_info->thread, SIGUSR1);
+	sem_post_exit_thread (conn_info);
 }
 
 static int conn_info_create (int fd)
@@ -769,6 +904,7 @@ static int conn_info_create (int fd)
 	memset (conn_info, 0, sizeof (struct conn_info));
 
 	conn_info->fd = fd;
+	conn_info->client_pid = 0;
 	conn_info->service = SOCKET_SERVICE_INIT;
 	conn_info->state = CONN_STATE_THREAD_INACTIVE;
 	list_init (&conn_info->outq_head);
@@ -776,7 +912,7 @@ static int conn_info_create (int fd)
 	list_init (&conn_info->zcb_mapped_list_head);
 	list_add (&conn_info->list, &conn_info_list_head);
 
-        api->poll_dispatch_add (fd, conn_info);
+	api->poll_dispatch_add (fd, conn_info);
 
 	return (0);
 }
@@ -793,15 +929,59 @@ static int conn_info_create (int fd)
 /*
  * Exported functions
  */
+extern void coroipcs_ipc_init_v2 (
+	struct coroipcs_init_state_v2 *init_state_v2)
+{
+	api = init_state_v2;
+	api->old_log_printf	= NULL;
+
+	log_printf (LOGSYS_LEVEL_DEBUG, "you are using ipc api v2\n");
+	_corosync_ipc_init ();
+}
+
 extern void coroipcs_ipc_init (
         struct coroipcs_init_state *init_state)
+{
+	api = calloc (sizeof(struct coroipcs_init_state_v2), 1);
+	/* v2 api */
+	api->stats_create_connection	= dummy_stats_create_connection;
+	api->stats_destroy_connection	= dummy_stats_destroy_connection;
+	api->stats_update_value		= dummy_stats_update_value;
+	api->stats_increment_value	= dummy_stats_increment_value;
+	api->log_printf			= NULL;
+
+	/* v1 api */
+	api->socket_name		= init_state->socket_name;
+	api->sched_policy		= init_state->sched_policy;
+	api->sched_param		= init_state->sched_param;
+	api->malloc			= init_state->malloc;
+	api->free			= init_state->free;
+	api->old_log_printf		= init_state->log_printf;
+	api->fatal_error		= init_state->fatal_error;
+	api->security_valid		= init_state->security_valid;
+	api->service_available		= init_state->service_available;
+	api->private_data_size_get	= init_state->private_data_size_get;
+	api->serialize_lock		= init_state->serialize_lock;
+	api->serialize_unlock		= init_state->serialize_unlock;
+	api->sending_allowed		= init_state->sending_allowed;
+	api->sending_allowed_release	= init_state->sending_allowed_release;
+	api->poll_accept_add		= init_state->poll_accept_add;
+	api->poll_dispatch_add		= init_state->poll_dispatch_add;
+	api->poll_dispatch_modify	= init_state->poll_dispatch_modify;
+	api->init_fn_get		= init_state->init_fn_get;
+	api->exit_fn_get		= init_state->exit_fn_get;
+	api->handler_fn_get		= init_state->handler_fn_get;
+
+	log_printf (LOGSYS_LEVEL_DEBUG, "you are using ipc api v1\n");
+
+	_corosync_ipc_init ();
+}
+
+static void _corosync_ipc_init(void)
 {
 	int server_fd;
 	struct sockaddr_un un_addr;
 	int res;
-
-	api = init_state;
-
 	/*
 	 * Create socket for IPC clients, name socket, listen for connections
 	 */
@@ -811,13 +991,13 @@ extern void coroipcs_ipc_init (
 	server_fd = socket (PF_LOCAL, SOCK_STREAM, 0);
 #endif
 	if (server_fd == -1) {
-		api->log_printf ("Cannot create client connections socket.\n");
+		log_printf (LOGSYS_LEVEL_CRIT, "Cannot create client connections socket.\n");
 		api->fatal_error ("Can't create library listen socket");
 	};
 
 	res = fcntl (server_fd, F_SETFL, O_NONBLOCK);
 	if (res == -1) {
-		api->log_printf ("Could not set non-blocking operation on server socket: %s\n", strerror (errno));
+		log_printf (LOGSYS_LEVEL_CRIT, "Could not set non-blocking operation on server socket: %s\n", strerror (errno));
 		api->fatal_error ("Could not set non-blocking operation on server socket");
 	}
 
@@ -834,7 +1014,7 @@ extern void coroipcs_ipc_init (
 		struct stat stat_out;
 		res = stat (SOCKETDIR, &stat_out);
 		if (res == -1 || (res == 0 && !S_ISDIR(stat_out.st_mode))) {
-			api->log_printf ("Required directory not present %s\n", SOCKETDIR);
+			log_printf (LOGSYS_LEVEL_CRIT, "Required directory not present %s\n", SOCKETDIR);
 			api->fatal_error ("Please create required directory.");
 		}
 		sprintf (un_addr.sun_path, "%s/%s", SOCKETDIR, api->socket_name);
@@ -844,7 +1024,7 @@ extern void coroipcs_ipc_init (
 
 	res = bind (server_fd, (struct sockaddr *)&un_addr, COROSYNC_SUN_LEN(&un_addr));
 	if (res) {
-		api->log_printf ("Could not bind AF_UNIX (%s): %s.\n", un_addr.sun_path, strerror (errno));
+		log_printf (LOGSYS_LEVEL_CRIT, "Could not bind AF_UNIX (%s): %s.\n", un_addr.sun_path, strerror (errno));
 		api->fatal_error ("Could not bind to AF_UNIX socket\n");
 	}
 
@@ -857,10 +1037,10 @@ extern void coroipcs_ipc_init (
 #endif
 	listen (server_fd, SERVER_BACKLOG);
 
-        /*
-         * Setup connection dispatch routine
-         */
-        api->poll_accept_add (server_fd);
+	/*
+	 * Setup connection dispatch routine
+	 */
+	api->poll_accept_add (server_fd);
 }
 
 void coroipcs_ipc_exit (void)
@@ -874,6 +1054,14 @@ void coroipcs_ipc_exit (void)
 
 		conn_info = list_entry (list, struct conn_info, list);
 
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+		sem_destroy (&conn_info->control_buffer->sem0);
+		sem_destroy (&conn_info->control_buffer->sem1);
+		sem_destroy (&conn_info->control_buffer->sem2);
+#else
+		semctl (conn_info->semid, 0, IPC_RMID);
+#endif
+
 		/*
 		 * Unmap memory segments
 		 */
@@ -886,9 +1074,7 @@ void coroipcs_ipc_exit (void)
 		res = circular_memory_unmap (conn_info->dispatch_buffer,
 			conn_info->dispatch_size);
 
-		semctl (conn_info->semid, 0, IPC_RMID);
-
-		pthread_kill (conn_info->thread, SIGUSR1);
+		sem_post_exit_thread (conn_info);
 	}
 }
 
@@ -905,10 +1091,19 @@ void *coroipcs_private_data_get (void *conn)
 int coroipcs_response_send (void *conn, const void *msg, size_t mlen)
 {
 	struct conn_info *conn_info = (struct conn_info *)conn;
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	struct sembuf sop;
+#endif
 	int res;
 
 	memcpy (conn_info->response_buffer, msg, mlen);
+
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+	res = sem_post (&conn_info->control_buffer->sem1);
+	if (res == -1) {
+		return (-1);
+	}
+#else
 	sop.sem_num = 1;
 	sop.sem_op = 1;
 	sop.sem_flg = 0;
@@ -916,18 +1111,23 @@ int coroipcs_response_send (void *conn, const void *msg, size_t mlen)
 retry_semop:
 	res = semop (conn_info->semid, &sop, 1);
 	if ((res == -1) && (errno == EINTR || errno == EAGAIN)) {
+		api->stats_increment_value (conn_info->stats_handle, "sem_retry_count");
 		goto retry_semop;
 	} else
 	if ((res == -1) && (errno == EINVAL || errno == EIDRM)) {
 		return (0);
 	}
+#endif
+	api->stats_increment_value (conn_info->stats_handle, "responses");
 	return (0);
 }
 
 int coroipcs_response_iov_send (void *conn, const struct iovec *iov, unsigned int iov_len)
 {
 	struct conn_info *conn_info = (struct conn_info *)conn;
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	struct sembuf sop;
+#endif
 	int res;
 	int write_idx = 0;
 	int i;
@@ -938,6 +1138,12 @@ int coroipcs_response_iov_send (void *conn, const struct iovec *iov, unsigned in
 		write_idx += iov[i].iov_len;
 	}
 
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+	res = sem_post (&conn_info->control_buffer->sem1);
+	if (res == -1) {
+		return (-1);
+	}
+#else
 	sop.sem_num = 1;
 	sop.sem_op = 1;
 	sop.sem_flg = 0;
@@ -945,11 +1151,14 @@ int coroipcs_response_iov_send (void *conn, const struct iovec *iov, unsigned in
 retry_semop:
 	res = semop (conn_info->semid, &sop, 1);
 	if ((res == -1) && (errno == EINTR || errno == EAGAIN)) {
+		api->stats_increment_value (conn_info->stats_handle, "sem_retry_count");
 		goto retry_semop;
 	} else
 	if ((res == -1) && (errno == EINVAL || errno == EIDRM)) {
 		return (0);
 	}
+#endif
+	api->stats_increment_value (conn_info->stats_handle, "responses");
 	return (0);
 }
 
@@ -980,21 +1189,55 @@ static void memcpy_dwrap (struct conn_info *conn_info, void *msg, unsigned int l
 	conn_info->control_buffer->write = (write_idx + len) % conn_info->dispatch_size;
 }
 
+/**
+ * simulate the behaviour in coroipcc.c
+ */
+static int flow_control_event_send (struct conn_info *conn_info, char event)
+{
+	int new_fc = 0;
+
+	if (event == MESSAGE_RES_OUTQ_NOT_EMPTY ||
+		event == MESSAGE_RES_ENABLE_FLOWCONTROL) {
+		new_fc = 1;
+	}
+
+	if (conn_info->flow_control_state != new_fc) {
+		if (new_fc == 1) {
+			log_printf (LOGSYS_LEVEL_INFO, "Enabling flow control for %d, event %d\n",
+				conn_info->client_pid, event);
+		} else {
+			log_printf (LOGSYS_LEVEL_INFO, "Disabling flow control for %d, event %d\n",
+				conn_info->client_pid, event);
+		}
+		conn_info->flow_control_state = new_fc;
+		api->stats_update_value (conn_info->stats_handle, "flow_control",
+			&conn_info->flow_control_state,
+			sizeof(conn_info->flow_control_state));
+		api->stats_increment_value (conn_info->stats_handle, "flow_control_count");
+	}
+
+	return send (conn_info->fd, &event, 1, MSG_NOSIGNAL);
+}
+
 static void msg_send (void *conn, const struct iovec *iov, unsigned int iov_len,
 		      int locked)
 {
 	struct conn_info *conn_info = (struct conn_info *)conn;
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	struct sembuf sop;
+#endif
 	int res;
 	int i;
-	char buf;
 
 	for (i = 0; i < iov_len; i++) {
 		memcpy_dwrap (conn_info, iov[i].iov_base, iov[i].iov_len);
 	}
 
-	buf = !list_empty (&conn_info->outq_head);
-	res = send (conn_info->fd, &buf, 1, MSG_NOSIGNAL);
+	if (list_empty (&conn_info->outq_head))
+		res = flow_control_event_send (conn_info, MESSAGE_RES_OUTQ_EMPTY);
+	else
+		res = flow_control_event_send (conn_info, MESSAGE_RES_OUTQ_NOT_EMPTY);
+
 	if (res == -1 && errno == EAGAIN) {
 		if (locked == 0) {
 			pthread_mutex_lock (&conn_info->mutex);
@@ -1009,6 +1252,9 @@ static void msg_send (void *conn, const struct iovec *iov, unsigned int iov_len,
 	if (res == -1) {
 		ipc_disconnect (conn_info);
 	}
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+	res = sem_post (&conn_info->control_buffer->sem2);
+#else
 	sop.sem_num = 2;
 	sop.sem_op = 1;
 	sop.sem_flg = 0;
@@ -1016,11 +1262,14 @@ static void msg_send (void *conn, const struct iovec *iov, unsigned int iov_len,
 retry_semop:
 	res = semop (conn_info->semid, &sop, 1);
 	if ((res == -1) && (errno == EINTR || errno == EAGAIN)) {
+		api->stats_increment_value (conn_info->stats_handle, "sem_retry_count");
 		goto retry_semop;
 	} else
 	if ((res == -1) && (errno == EINVAL || errno == EIDRM)) {
 		return;
 	}
+#endif
+	api->stats_increment_value (conn_info->stats_handle, "dispatched");
 }
 
 static void outq_flush (struct conn_info *conn_info) {
@@ -1028,13 +1277,11 @@ static void outq_flush (struct conn_info *conn_info) {
 	struct outq_item *outq_item;
 	unsigned int bytes_left;
 	struct iovec iov;
-	char buf;
 	int res;
 
 	pthread_mutex_lock (&conn_info->mutex);
 	if (list_empty (&conn_info->outq_head)) {
-		buf = 3;
-		res = send (conn_info->fd, &buf, 1, MSG_NOSIGNAL);
+		res = flow_control_event_send (conn_info, MESSAGE_RES_OUTQ_FLUSH_NR);
 		pthread_mutex_unlock (&conn_info->mutex);
 		return;
 	}
@@ -1051,6 +1298,7 @@ static void outq_flush (struct conn_info *conn_info) {
 			list_del (list);
 			api->free (iov.iov_base);
 			api->free (outq_item);
+			api->stats_decrement_value (conn_info->stats_handle, "queue_size");
 		} else {
 			break;
 		}
@@ -1062,18 +1310,22 @@ static int priv_change (struct conn_info *conn_info)
 {
 	mar_req_priv_change req_priv_change;
 	unsigned int res;
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	union semun semun;
 	struct semid_ds ipc_set;
 	int i;
+#endif
 
 retry_recv:
 	res = recv (conn_info->fd, &req_priv_change,
 		sizeof (mar_req_priv_change),
 		MSG_NOSIGNAL);
 	if (res == -1 && errno == EINTR) {
+		api->stats_increment_value (conn_info->stats_handle, "recv_retry_count");
 		goto retry_recv;
 	}
 	if (res == -1 && errno == EAGAIN) {
+		api->stats_increment_value (conn_info->stats_handle, "recv_retry_count");
 		goto retry_recv;
 	}
 	if (res == -1 && errno != EAGAIN) {
@@ -1087,6 +1339,7 @@ retry_recv:
 	}
 #endif
 
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	ipc_set.sem_perm.uid = req_priv_change.euid;
 	ipc_set.sem_perm.gid = req_priv_change.egid;
 	ipc_set.sem_perm.mode = 0600;
@@ -1099,6 +1352,7 @@ retry_recv:
 			return (-1);
 		}
 	}
+#endif
 	return (0);
 }
 
@@ -1150,6 +1404,7 @@ static void msg_send_or_queue (void *conn, const struct iovec *iov, unsigned int
 		}
 		list_add_tail (&outq_item->list, &conn_info->outq_head);
 		pthread_mutex_unlock (&conn_info->mutex);
+		api->stats_increment_value (conn_info->stats_handle, "queue_size");
 		return;
 	}
 	msg_send (conn, iov, iov_len, MSG_SEND_LOCKED);
@@ -1212,13 +1467,16 @@ retry_accept:
 	}
 
 	if (new_fd == -1) {
-		api->log_printf ("Could not accept Library connection: %s\n", strerror (errno));
+		log_printf (LOGSYS_LEVEL_ERROR,
+			"Could not accept Library connection: %s\n", strerror (errno));
 		return (0); /* This is an error, but -1 would indicate disconnect from poll loop */
 	}
 
 	res = fcntl (new_fd, F_SETFL, O_NONBLOCK);
 	if (res == -1) {
-		api->log_printf ("Could not set non-blocking operation on library connection: %s\n", strerror (errno));
+		log_printf (LOGSYS_LEVEL_ERROR,
+			"Could not set non-blocking operation on library connection: %s\n",
+			strerror (errno));
 		close (new_fd);
 		return (0); /* This is an error, but -1 would indicate disconnect from poll loop */
 	}
@@ -1242,6 +1500,69 @@ retry_accept:
 	return (0);
 }
 
+static char * pid_to_name (pid_t pid, char *out_name, size_t name_len)
+{
+	char *name;
+	char *rest;
+	FILE *fp;
+	char fname[32];
+	char buf[256];
+
+	snprintf (fname, 32, "/proc/%d/stat", pid);
+	fp = fopen (fname, "r");
+	if (!fp) {
+		return NULL;
+	}
+
+	if (fgets (buf, sizeof (buf), fp) == NULL) {
+		fclose (fp);
+		return NULL;
+	}
+	fclose (fp);
+
+	name = strrchr (buf, '(');
+	if (!name) {
+		return NULL;
+	}
+
+	/* move past the bracket */
+	name++;
+
+	rest = strrchr (buf, ')');
+
+	if (rest == NULL || rest[1] != ' ') {
+		return NULL;
+	}
+
+	*rest = '\0';
+	/* move past the NULL and space */
+	rest += 2;
+
+	/* copy the name */
+	strncpy (out_name, name, name_len);
+	out_name[name_len - 1] = '\0';
+	return out_name;
+}
+
+static void coroipcs_init_conn_stats (
+	struct conn_info *conn)
+{
+	char conn_name[42];
+	char proc_name[32];
+
+	if (conn->client_pid > 0) {
+		if (pid_to_name (conn->client_pid, proc_name, sizeof(proc_name)))
+			snprintf (conn_name, sizeof(conn_name), "%s:%d:%d", proc_name, conn->client_pid, conn->fd);
+		else
+			snprintf (conn_name, sizeof(conn_name), "%d:%d", conn->client_pid, conn->fd);
+	} else
+		snprintf (conn_name, sizeof(conn_name), "%d", conn->fd);
+
+	conn->stats_handle = api->stats_create_connection (conn_name, conn->client_pid, conn->fd);
+	api->stats_update_value (conn->stats_handle, "service_id",
+		&conn->service, sizeof(conn->service));
+}
+
 int coroipcs_handler_dispatch (
 	int fd,
 	int revent,
@@ -1281,7 +1602,6 @@ int coroipcs_handler_dispatch (
 		if (res != 1) {
 			return (0);
 		}
-		req_setup_send (conn_info, CS_OK);
 
 		pthread_mutex_init (&conn_info->mutex, NULL);
 		req_setup = (mar_req_setup_t *)conn_info->setup_msg;
@@ -1289,11 +1609,15 @@ int coroipcs_handler_dispatch (
 		 * Is the service registered ?
 		 */
 		if (api->service_available (req_setup->service) == 0) {
+			req_setup_send (conn_info, CS_ERR_NOT_EXIST);
 			ipc_disconnect (conn_info);
 			return (0);
 		}
+		req_setup_send (conn_info, CS_OK);
 
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 		conn_info->semkey = req_setup->semkey;
+#endif
 		res = memory_map (
 			req_setup->control_file,
 			req_setup->control_size,
@@ -1323,7 +1647,9 @@ int coroipcs_handler_dispatch (
 		conn_info->notify_flow_control_enabled = 0;
 		conn_info->setup_bytes_read = 0;
 
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 		conn_info->semid = semget (conn_info->semkey, 3, 0600);
+#endif
 		conn_info->pending_semops = 0;
 
 		/*
@@ -1338,6 +1664,9 @@ int coroipcs_handler_dispatch (
 
 		api->init_fn_get (conn_info->service) (conn_info);
 
+		/* create stats objects */
+		coroipcs_init_conn_stats (conn_info);
+
 		pthread_attr_init (&conn_info->thread_attr);
 		/*
 		* IA64 needs more stack space then other arches
@@ -1396,9 +1725,13 @@ int coroipcs_handler_dispatch (
 	coroipcs_refcount_inc (conn_info);
 	pthread_mutex_lock (&conn_info->mutex);
 	if ((conn_info->state == CONN_STATE_THREAD_ACTIVE) && (revent & POLLOUT)) {
-		buf = !list_empty (&conn_info->outq_head);
+		if (list_empty (&conn_info->outq_head))
+			buf = MESSAGE_RES_OUTQ_EMPTY;
+		else
+			buf = MESSAGE_RES_OUTQ_NOT_EMPTY;
+
 		for (; conn_info->pending_semops;) {
-			res = send (conn_info->fd, &buf, 1, MSG_NOSIGNAL);
+			res = flow_control_event_send (conn_info, buf);
 			if (res == 1) {
 				conn_info->pending_semops--;
 			} else {
@@ -1406,8 +1739,7 @@ int coroipcs_handler_dispatch (
 			}
 		}
 		if (conn_info->notify_flow_control_enabled) {
-			buf = 2;
-			res = send (conn_info->fd, &buf, 1, MSG_NOSIGNAL);
+			res = flow_control_event_send (conn_info, MESSAGE_RES_ENABLE_FLOWCONTROL);
 			if (res == 1) {
 				conn_info->notify_flow_control_enabled = 0;
 			}
@@ -1424,3 +1756,4 @@ int coroipcs_handler_dispatch (
 
 	return (0);
 }
+
