@@ -131,6 +131,10 @@ unsigned int flt_data_size;
 
 static int logsys_system_needs_init = LOGSYS_LOGGER_NEEDS_INIT;
 
+static int logsys_sched_param_queued = 0;
+static int logsys_sched_policy;
+static struct sched_param logsys_sched_param;
+
 static int logsys_after_log_ops_yield = 10;
 
 /*
@@ -406,6 +410,11 @@ static inline int strcpy_cutoff (char *dest, const char *src, size_t cutoff,
 	return (cutoff);
 }
 
+static const char log_month_name[][4] = {
+	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
+	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
+};
+
 /*
  * %s SUBSYSTEM
  * %n FUNCTION NAME
@@ -437,6 +446,7 @@ static void log_printf_to_logs (
 	int subsysid;
 	unsigned int level;
 	int c;
+	struct tm tm_res;
 
 	if (LOGSYS_DECODE_RECID(rec_ident) != LOGSYS_RECID_LOG) {
 		return;
@@ -486,7 +496,10 @@ static void log_printf_to_logs (
 
 				case 't':
 					gettimeofday (&tv, NULL);
-					(void)strftime (char_time, sizeof (char_time), "%b %d %T", localtime ((time_t *)&tv.tv_sec));
+					(void)localtime_r ((time_t *)&tv.tv_sec, &tm_res);
+					snprintf (char_time, sizeof (char_time), "%s %02d %02d:%02d:%02d",
+					    log_month_name[tm_res.tm_mon], tm_res.tm_mday, tm_res.tm_hour,
+					    tm_res.tm_min, tm_res.tm_sec);
 					normal_p = char_time;
 
 					/*
@@ -523,8 +536,12 @@ static void log_printf_to_logs (
 			syslog_output_buffer_idx += syslog_len;
 			format_buffer_idx += 1;
 		}
-		if ((normal_output_buffer_idx == sizeof (normal_output_buffer)) ||
-		    (syslog_output_buffer_idx == sizeof (syslog_output_buffer))) {
+		if ((normal_output_buffer_idx >= sizeof (normal_output_buffer) - 2) ||
+		    (syslog_output_buffer_idx >= sizeof (syslog_output_buffer) - 1)) {
+			/* Note: we make allowance for '\0' at the end of
+			 * both of these arrays and normal_output_buffer also
+			 * needs a '\n'.
+			 */
 			break;
 		}
 	}
@@ -783,14 +800,31 @@ static void wthread_create (void)
 	pthread_mutex_init (&logsys_cond_mutex, NULL);
 	pthread_cond_init (&logsys_cond, NULL);
 	pthread_mutex_lock (&logsys_cond_mutex);
+
+	/*
+	 * TODO: propagate pthread_create errors back to the caller
+	 */
 	res = pthread_create (&logsys_thread_id, NULL,
 		logsys_worker_thread, NULL);
 
-
-	/*
-	 * Wait for thread to be started
-	 */
-	wthread_wait_locked ();
+	if (res == 0) {
+		/*
+		 * Wait for thread to be started
+		 */
+		wthread_wait_locked ();
+		if (logsys_sched_param_queued == 1) {
+			/*
+			 * TODO: propagate logsys_thread_priority_set errors back to
+			 * the caller
+			 */
+			res = logsys_thread_priority_set (logsys_sched_policy,
+						    &logsys_sched_param,
+						    logsys_after_log_ops_yield);
+			logsys_sched_param_queued = 0;
+		}
+	} else {
+		wthread_active = 0;
+	}
 }
 
 static int _logsys_config_subsys_get_unlocked (const char *subsys)
@@ -803,7 +837,6 @@ static int _logsys_config_subsys_get_unlocked (const char *subsys)
 
  	for (i = 0; i <= LOGSYS_MAX_SUBSYS_COUNT; i++) {
 		if (strcmp (logsys_loggers[i].subsys, subsys) == 0) {
-			pthread_mutex_unlock (&logsys_config_mutex);
 			return i;
 		}
 	}
@@ -1283,6 +1316,7 @@ void _logsys_log_vprintf (
 	unsigned int len;
 	unsigned int level;
 	int subsysid;
+	const char *short_file_name;
 
 	subsysid = LOGSYS_DECODE_SUBSYSID(rec_ident);
 	level = LOGSYS_DECODE_LEVEL(rec_ident);
@@ -1298,6 +1332,15 @@ void _logsys_log_vprintf (
 		logsys_print_buffer[len - 1] = '\0';
 		len -= 1;
 	}
+#ifdef BUILDING_IN_PLACE
+	short_file_name = file_name;
+#else
+	short_file_name = strrchr (file_name, '/');
+	if (short_file_name == NULL)
+		short_file_name = file_name;
+	else
+		short_file_name++; /* move past the "/" */
+#endif /* BUILDING_IN_PLACE */
 
 	/*
 	 * Create a log record
@@ -1305,7 +1348,7 @@ void _logsys_log_vprintf (
 	_logsys_log_rec (
 		rec_ident,
 		function_name,
-		file_name,
+		short_file_name,
 		file_line,
 		logsys_print_buffer, len + 1,
 		LOGSYS_REC_END);
@@ -1316,7 +1359,7 @@ void _logsys_log_vprintf (
 		 * expect the worker thread to output the log data once signaled
 		 */
 		log_printf_to_logs (rec_ident,
-				    file_name, function_name, file_line,
+				    short_file_name, function_name, file_line,
 				    logsys_print_buffer);
 	} else {
 		/*
@@ -1610,15 +1653,20 @@ int logsys_thread_priority_set (
 	int res = 0;
 
 #if defined(HAVE_PTHREAD_SETSCHEDPARAM) && defined(HAVE_SCHED_GET_PRIORITY_MAX)
-	if (policy != SCHED_OTHER) {
+	if (wthread_active == 0) {
+		logsys_sched_policy = policy;
+		memcpy(&logsys_sched_param, &param, sizeof(struct sched_param));
+		logsys_sched_param_queued = 1;
+	} else {
 		res = pthread_setschedparam (logsys_thread_id, policy, param);
 	}
 #endif
+
 	if (after_log_ops_yield > 0) {
 		logsys_after_log_ops_yield = after_log_ops_yield;
 	}
-	return (res);
 
+	return (res);
 }
 
 int logsys_log_rec_store (const char *filename)
