@@ -35,6 +35,7 @@
 #ifndef COROSYNC_SERVICE_H_DEFINED
 #define COROSYNC_SERVICE_H_DEFINED
 
+#include <corosync/hdb.h>
 /*
  * Link and initialize a service
  */
@@ -45,13 +46,6 @@ extern unsigned int corosync_service_link_and_init (
 	const char *service_name,
 	unsigned int service_ver);
 
-/*
- * Unlink and exit a service based on service priority
- */
-extern unsigned int corosync_service_unlink_priority (
-	struct corosync_api_v1 *corosync_api,
-	int priority);
-
 /*
  * Unlink and exit a service
  */
@@ -63,8 +57,9 @@ extern unsigned int corosync_service_unlink_and_exit (
 /*
  * Unlink and exit all corosync services
  */
-extern unsigned int corosync_service_unlink_all (
-	struct corosync_api_v1 *objdb);
+extern void corosync_service_unlink_all (
+        struct corosync_api_v1 *api,
+        void (*unlink_all_complete) (void));
 
 /*
  * Load all of the default services
@@ -74,4 +69,6 @@ extern unsigned int corosync_service_defaults_link_and_init (
 
 extern struct corosync_service_engine *ais_service[];
 
+extern hdb_handle_t service_stats_handle[SERVICE_HANDLER_MAXIMUM_COUNT][64];
+
 #endif /* SERVICE_H_DEFINED */
