@@ -71,7 +71,6 @@
 #define TOKEN_RETRANSMIT_TIMEOUT		(int)(TOKEN_TIMEOUT / (TOKEN_RETRANSMITS_BEFORE_LOSS_CONST + 0.2))
 #define TOKEN_HOLD_TIMEOUT			(int)(TOKEN_RETRANSMIT_TIMEOUT * 0.8 - (1000/(int)HZ))
 #define JOIN_TIMEOUT				50
-#define CONSENSUS_TIMEOUT			800
 #define MERGE_TIMEOUT				200
 #define DOWNCHECK_TIMEOUT			1000
 #define FAIL_TO_RECV_CONST			50
@@ -220,7 +219,6 @@ static void totem_volatile_config_read (
 	objdb_get_string (objdb, object_totem_handle, "vsftype", &totem_config->vsf_type);
 
 	objdb_get_int (objdb,object_totem_handle, "max_messages", &totem_config->max_messages);
-
 }
 
 
@@ -275,6 +273,7 @@ extern int totem_config_read (
 	const char *str;
 	unsigned int ringnumber = 0;
 	hdb_handle_t object_find_interface_handle;
+	const char *transport_type;
 
 	res = totem_handle_find (objdb, &object_totem_handle);
 	if (res == -1) {
@@ -392,6 +391,14 @@ printf ("couldn't find totem handle\n");
 
 	add_totem_config_notification(objdb, totem_config, object_totem_handle);
 
+	totem_config->transport_number = 0;
+	objdb_get_string (objdb, object_totem_handle, "transport", &transport_type);
+	if (transport_type) {
+		if (strcmp (transport_type, "iba") == 0) {
+			totem_config->transport_number = 1;
+		}
+	}
+
 	return 0;
 }
 
@@ -542,7 +549,7 @@ int totem_config_validate (
 	}
 
 	if (totem_config->consensus_timeout == 0) {
-		totem_config->consensus_timeout = CONSENSUS_TIMEOUT;
+		totem_config->consensus_timeout = (int)(float)(1.2 * totem_config->token_timeout);
 	}
 
 	if (totem_config->consensus_timeout < MINIMUM_TIMEOUT) {
@@ -552,6 +559,13 @@ int totem_config_validate (
 		goto parse_error;
 	}
 
+	if (totem_config->consensus_timeout <= 1.2 * totem_config->token_timeout) {
+		snprintf (local_error_reason, sizeof(local_error_reason),
+			"The consensus timeout parameter (%d ms) must be atleast 1.2 * token (%d ms).",
+			totem_config->consensus_timeout, (int) ((float)1.2 * totem_config->token_timeout));
+		goto parse_error;
+	}
+
 	if (totem_config->merge_timeout == 0) {
 		totem_config->merge_timeout = MERGE_TIMEOUT;
 	}
