@@ -65,16 +65,17 @@ struct corosync_tpg_group {
 
 #define TOTEMIP_ADDRLEN (sizeof(struct in6_addr))
 
-#define PROCESSOR_COUNT_MAX 384
 #define INTERFACE_MAX 2
 
-#ifndef MESSAGE_SIZE_MAX
+#ifdef HAVE_SMALL_MEMORY_FOOTPRINT
+#define PROCESSOR_COUNT_MAX	16
+#define MESSAGE_SIZE_MAX	1024*64
+#define MESSAGE_QUEUE_MAX	512
+#else
+#define PROCESSOR_COUNT_MAX	384
 #define MESSAGE_SIZE_MAX	1024*1024 /* (1MB) */
-#endif /* MESSAGE_SIZE_MAX */
-
-#ifndef MESSAGE_QUEUE_MAX
-#define MESSAGE_QUEUE_MAX MESSAGE_SIZE_MAX / totem_config->net_mtu
-#endif /* MESSAGE_QUEUE_MAX */
+#define MESSAGE_QUEUE_MAX	MESSAGE_SIZE_MAX / totem_config->net_mtu
+#endif /* HAVE_SMALL_MEMORY_FOOTPRINT */
 
 #define TOTEM_AGREED	0
 #define TOTEM_SAFE	1
@@ -168,6 +169,19 @@ struct object_key_valid {
 };
 /* deprecated */
 
+typedef enum {
+	OBJDB_VALUETYPE_INT16,
+	OBJDB_VALUETYPE_UINT16,
+	OBJDB_VALUETYPE_INT32,
+	OBJDB_VALUETYPE_UINT32,
+	OBJDB_VALUETYPE_INT64,
+	OBJDB_VALUETYPE_UINT64,
+	OBJDB_VALUETYPE_FLOAT,
+	OBJDB_VALUETYPE_DOUBLE,
+	OBJDB_VALUETYPE_STRING,
+	OBJDB_VALUETYPE_ANY,
+} objdb_value_types_t;
+
 typedef enum {
 	OBJECT_TRACK_DEPTH_ONE,
 	OBJECT_TRACK_DEPTH_RECURSIVE
@@ -589,6 +603,30 @@ struct corosync_api_v1 {
 	 * Please avoid using any of coropoll apis in your service engines.
 	 */
 	hdb_handle_t (*poll_handle_get) (void);
+
+
+	int (*object_key_create_typed) (
+		hdb_handle_t object_handle,
+		const char *key_name,
+		const void *value,
+		size_t value_len,
+		objdb_value_types_t type);
+
+	int (*object_key_get_typed) (
+		hdb_handle_t object_handle,
+		const char *key_name,
+		void **value,
+		size_t *value_len,
+		objdb_value_types_t *type);
+
+	int (*object_key_iter_typed) (
+		hdb_handle_t parent_object_handle,
+		char **key_name,
+		void **value,
+		size_t *value_len,
+		objdb_value_types_t *type);
+
+	void *(*totem_get_stats)(void);
 };
 
 #define SERVICE_ID_MAKE(a,b) ( ((a)<<16) | (b) )
