@@ -2,7 +2,7 @@
  * Copyright (C) 2006 Steven Dake (sdake@redhat.com)
  *
  * This software licensed under BSD license, the text of which follows:
- * 
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *
@@ -60,7 +60,12 @@ struct lcr_iface_instance {
 	void *context;
 	void (*destructor) (void *context);
 };
-	
+
+DECLARE_HDB_DATABASE (lcr_component_instance_database, NULL);
+
+DECLARE_HDB_DATABASE (lcr_iface_instance_database, NULL);
+
+/*
 static struct hdb_handle_database lcr_component_instance_database = {
 	.handle_count	= 0,
 	.handles	= 0,
@@ -72,6 +77,7 @@ static struct hdb_handle_database lcr_iface_instance_database = {
 	.handles	= 0,
 	.iterator	= 0
 };
+*/
 
 static hdb_handle_t g_component_handle = 0xFFFFFFFF;
 
@@ -93,8 +99,7 @@ static int lcr_select_so (struct dirent *dirent)
 	return (0);
 }
 
-#ifndef COROSYNC_SOLARIS
-#ifdef COROSYNC_LINUX
+#if defined(COROSYNC_LINUX) || defined(COROSYNC_SOLARIS)
 static int pathlist_select (const struct dirent *dirent)
 #else
 static int pathlist_select (struct dirent *dirent)
@@ -106,7 +111,6 @@ static int pathlist_select (struct dirent *dirent)
 
 	return (0);
 }
-#endif
 
 static inline struct lcr_component_instance *lcr_comp_find (
 	const char *iface_name,
@@ -180,7 +184,7 @@ static void defaults_path_build (void)
 	if (res != NULL && (path_list[0] = strdup (cwd)) != NULL) {
 		path_list_entries++;
 	}
-	
+
 	path_list[path_list_entries++] = LCRSODIR;
 }
 
@@ -213,7 +217,6 @@ static void ld_library_path_build (void)
 
 static int ldso_path_build (const char *path, const char *filename)
 {
-#ifndef COROSYNC_SOLARIS
 	FILE *fp;
 	char string[1024];
 	char filename_cat[1024];
@@ -224,7 +227,7 @@ static int ldso_path_build (const char *path, const char *filename)
 	struct dirent **scandir_list;
 	unsigned int scandir_entries;
 
-	sprintf (filename_cat, "%s/%s", path, filename);
+	snprintf (filename_cat, sizeof(filename_cat), "%s/%s", path, filename);
 	if (filename[0] == '*') {
 		scandir_entries = scandir (
 			path,
@@ -272,7 +275,6 @@ static int ldso_path_build (const char *path, const char *filename)
 		}
 	}
 	fclose(fp);
-#endif
 	return (0);
 }
 
@@ -335,9 +337,7 @@ fail:
 		namelist_items--;
 		free (*namelist[namelist_items]);
 	}
-	if (names != NULL) {
-		free (names);
-	}
+	free (names);
 	*namelist = NULL;
 	errno = err;
 	return -1;
@@ -365,6 +365,9 @@ static int interface_find_and_load (
 	int scandir_entries;
 	unsigned int libs_to_scan;
 	char dl_name[1024];
+#ifdef COROSYNC_SOLARIS
+	void (*comp_reg)(void);
+#endif
 
 	scandir_entries = scandir (path,  &scandir_list, lcr_select_so, alphasort);
 	if (scandir_entries > 0)
@@ -375,19 +378,28 @@ static int interface_find_and_load (
 		/*
 		 * Load objects, scan them, unload them if they are not a match
 		 */
-		sprintf (dl_name, "%s/%s", path, scandir_list[libs_to_scan]->d_name);
+		snprintf (dl_name, sizeof(dl_name), "%s/%s",
+			path, scandir_list[libs_to_scan]->d_name);
 		/*
 	 	 * Don't reload already loaded libraries
 		 */
 		if (lcr_lib_loaded (dl_name)) {
 			continue;
 		}
-		dl_handle = dlopen (dl_name, RTLD_LAZY);
+		dl_handle = dlopen (dl_name, RTLD_NOW);
 		if (dl_handle == NULL) {
 			fprintf(stderr, "%s: open failed: %s\n",
 				dl_name, dlerror());
 			continue;
 		}
+/*
+ * constructors don't work in Solaris dlopen, so we have to specifically call
+ * a function to register the component
+ */
+#ifdef COROSYNC_SOLARIS
+		comp_reg = dlsym (dl_handle, "corosync_lcr_component_register");
+		comp_reg ();
+#endif
 		instance = lcr_comp_find (iface_name, version, iface_number);
 		if (instance) {
 			instance->dl_handle = dl_handle;
@@ -414,6 +426,7 @@ static int interface_find_and_load (
 		}
 		free (scandir_list);
 	}
+	g_component_handle = 0xFFFFFFFF;
 	return -1;
 
 found:
@@ -425,6 +438,7 @@ found:
 		}
 		free (scandir_list);
 	}
+	g_component_handle = 0xFFFFFFFF;
 	return 0;
 }
 
@@ -521,7 +535,6 @@ void lcr_component_register (struct lcr_comp *comp)
 {
 	struct lcr_component_instance *instance;
 	static hdb_handle_t comp_handle;
-	int res;
 
 	hdb_handle_create (&lcr_component_instance_database,
 		sizeof (struct lcr_component_instance),
