@@ -178,17 +178,17 @@ corosync_cfg_dispatch (
 			error = CS_OK;
 			goto error_put;
 		}
-		if (error != CS_OK) {
-			goto error_put;
-		}
-
-		if (dispatch_data == NULL) {
+		if (error == CS_ERR_TRY_AGAIN) {
+			error = CS_OK;
 			if (dispatch_flags == CPG_DISPATCH_ALL) {
 				break; /* exit do while cont is 1 loop */
 			} else {
 				continue; /* next poll */
 			}
 		}
+		if (error != CS_OK) {
+			goto error_put;
+		}
 
 		/*
 		 * Make copy of callbacks, message data, unlock instance, and call callback
@@ -202,10 +202,12 @@ corosync_cfg_dispatch (
 		 */
 		switch (dispatch_data->id) {
 		case MESSAGE_RES_CFG_TESTSHUTDOWN:
-			if (callbacks.corosync_cfg_shutdown_callback) {
-				res_lib_cfg_testshutdown = (struct res_lib_cfg_testshutdown *)dispatch_data;
-				callbacks.corosync_cfg_shutdown_callback(cfg_handle, res_lib_cfg_testshutdown->flags);
+			if (callbacks.corosync_cfg_shutdown_callback == NULL) {
+				continue;
 			}
+
+			res_lib_cfg_testshutdown = (struct res_lib_cfg_testshutdown *)dispatch_data;
+			callbacks.corosync_cfg_shutdown_callback(cfg_handle, res_lib_cfg_testshutdown->flags);
 			break;
 		default:
 			coroipcc_dispatch_put (cfg_instance->handle);
@@ -218,14 +220,8 @@ corosync_cfg_dispatch (
 		/*
 		 * Determine if more messages should be processed
 		 */
-		switch (dispatch_flags) {
-		case CS_DISPATCH_ONE:
+		if (dispatch_flags == CS_DISPATCH_ONE) {
 			cont = 0;
-			break;
-		case CS_DISPATCH_ALL:
-			break;
-		case CS_DISPATCH_BLOCKING:
-			break;
 		}
 	} while (cont);
 
