@@ -55,7 +55,6 @@
 #include <netinet/in.h>
 #include <assert.h>
 #include <sys/shm.h>
-#include <sys/sem.h>
 #include <sys/mman.h>
 
 #include <corosync/corotypes.h>
@@ -64,12 +63,19 @@
 #include <corosync/coroipcc.h>
 #include <corosync/hdb.h>
 
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+#include <semaphore.h>
+#else
+#include <sys/sem.h>
+#endif
+
 #include "util.h"
 
 struct ipc_instance {
 	int fd;
-	int shmid;
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	int semid;
+#endif
 	int flow_control_state;
 	struct control_buffer *control_buffer;
 	char *request_buffer;
@@ -226,6 +232,7 @@ res_exit:
 	return (res);
 }
 
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 static int
 priv_change_send (struct ipc_instance *ipc_instance)
 {
@@ -266,6 +273,7 @@ union semun {
         struct seminfo *__buf;
 };
 #endif
+#endif
 
 static int
 circular_memory_map (char *path, const char *file, void **buf, size_t bytes)
@@ -391,17 +399,31 @@ msg_send (
 	const struct iovec *iov,
 	unsigned int iov_len)
 {
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	struct sembuf sop;
+#endif
+
 	int i;
 	int res;
 	int req_buffer_idx = 0;
 
 	for (i = 0; i < iov_len; i++) {
+		if ((req_buffer_idx + iov[i].iov_len) > 
+			ipc_instance->request_size) {
+			return (CS_ERR_INVALID_PARAM);
+		}
 		memcpy (&ipc_instance->request_buffer[req_buffer_idx],
 			iov[i].iov_base,
 			iov[i].iov_len);
 		req_buffer_idx += iov[i].iov_len;
 	}
+
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+	res = sem_post (&ipc_instance->control_buffer->sem0);
+	if (res == -1) {
+		return (CS_ERR_LIBRARY);
+	}
+#else 
 	/*
 	 * Signal semaphore #0 indicting a new message from client
 	 * to server request queue
@@ -422,6 +444,7 @@ retry_semop:
 	if (res == -1) {
 		return (CS_ERR_LIBRARY);
 	}
+#endif
 	return (CS_OK);
 }
 
@@ -431,10 +454,19 @@ reply_receive (
 	void *res_msg,
 	size_t res_len)
 {
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	struct sembuf sop;
+#endif
 	coroipc_response_header_t *response_header;
 	int res;
 
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+retry_semwait:
+	res = sem_wait (&ipc_instance->control_buffer->sem1);
+	if (res == -1 && errno == EINTR) {
+		goto retry_semwait;
+	}
+#else
 	/*
 	 * Wait for semaphore #1 indicating a new message from server
 	 * to client in the response queue
@@ -455,6 +487,7 @@ retry_semop:
 	if (res == -1) {
 		return (CS_ERR_LIBRARY);
 	}
+#endif
 
 	response_header = (coroipc_response_header_t *)ipc_instance->response_buffer;
 	if (response_header->error == CS_ERR_TRY_AGAIN) {
@@ -470,9 +503,18 @@ reply_receive_in_buf (
 	struct ipc_instance *ipc_instance,
 	void **res_msg)
 {
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	struct sembuf sop;
+#endif
 	int res;
 
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+retry_semwait:
+	res = sem_wait (&ipc_instance->control_buffer->sem1);
+	if (res == -1 && errno == EINTR) {
+		goto retry_semwait;
+	}
+#else
 	/*
 	 * Wait for semaphore #1 indicating a new message from server
 	 * to client in the response queue
@@ -493,6 +535,7 @@ retry_semop:
 	if (res == -1) {
 		return (CS_ERR_LIBRARY);
 	}
+#endif
 
 	*res_msg = (char *)ipc_instance->response_buffer;
 	return (CS_OK);
@@ -515,11 +558,13 @@ coroipcc_service_connect (
 	struct sockaddr_un address;
 	cs_error_t res;
 	struct ipc_instance *ipc_instance;
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	key_t semkey = 0;
+	union semun semun;
+#endif
 	int sys_res;
 	mar_req_setup_t req_setup;
 	mar_res_setup_t res_setup;
-	union semun semun;
 	char control_map_path[128];
 	char request_map_path[128];
 	char response_map_path[128];
@@ -568,6 +613,36 @@ coroipcc_service_connect (
 		return (CS_ERR_TRY_AGAIN);
 	}
 
+	res = memory_map (
+		control_map_path,
+		"control_buffer-XXXXXX",
+		(void *)&ipc_instance->control_buffer,
+		8192);
+
+	res = memory_map (
+		request_map_path,
+		"request_buffer-XXXXXX",
+		(void *)&ipc_instance->request_buffer,
+		request_size);
+
+	res = memory_map (
+		response_map_path,
+		"response_buffer-XXXXXX",
+		(void *)&ipc_instance->response_buffer,
+		response_size);
+
+	res = circular_memory_map (
+		dispatch_map_path,
+		"dispatch_buffer-XXXXXX",
+		(void *)&ipc_instance->dispatch_buffer,
+		dispatch_size);
+
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+	sem_init (&ipc_instance->control_buffer->sem0, 1, 0);
+	sem_init (&ipc_instance->control_buffer->sem1, 1, 0);
+	sem_init (&ipc_instance->control_buffer->sem2, 1, 0);
+#else
+
 	/*
 	 * Allocate a semaphore segment
 	 */
@@ -600,30 +675,7 @@ coroipcc_service_connect (
 	if (res != 0) {
 		goto res_exit;
 	}
-
-	res = memory_map (
-		control_map_path,
-		"control_buffer-XXXXXX",
-		(void *)&ipc_instance->control_buffer,
-		8192);
-
-	res = memory_map (
-		request_map_path,
-		"request_buffer-XXXXXX",
-		(void *)&ipc_instance->request_buffer,
-		request_size);
-
-	res = memory_map (
-		response_map_path,
-		"response_buffer-XXXXXX",
-		(void *)&ipc_instance->response_buffer,
-		response_size);
-
-	res = circular_memory_map (
-		dispatch_map_path,
-		"dispatch_buffer-XXXXXX",
-		(void *)&ipc_instance->dispatch_buffer,
-		dispatch_size);
+#endif
 
 	/*
 	 * Initialize IPC setup message
@@ -637,7 +689,10 @@ coroipcc_service_connect (
 	req_setup.request_size = request_size;
 	req_setup.response_size = response_size;
 	req_setup.dispatch_size = dispatch_size;
+
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	req_setup.semkey = semkey;
+#endif
 
 	res = socket_send (request_fd, &req_setup, sizeof (mar_req_setup_t));
 	if (res != CS_OK) {
@@ -668,8 +723,10 @@ coroipcc_service_connect (
 
 res_exit:
 	close (request_fd);
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	if (ipc_instance->semid > 0)
 		semctl (ipc_instance->semid, 0, IPC_RMID);
+#endif
 	return (res_setup.error);
 }
 
@@ -762,9 +819,11 @@ coroipcc_dispatch_get (
 		goto error_put;
 	} else
 	if (poll_events == -1) {
+		error = CS_ERR_LIBRARY;
 		goto error_put;
 	} else
 	if (poll_events == 0) {
+		error = CS_ERR_TRY_AGAIN;
 		goto error_put;
 	}
 	if (poll_events == 1 && (ufds.revents & (POLLERR|POLLHUP))) {
@@ -774,17 +833,23 @@ coroipcc_dispatch_get (
 
 	res = recv (ipc_instance->fd, &buf, 1, 0);
 	if (res == -1 && errno == EINTR) {
-		res = CS_ERR_TRY_AGAIN;
+		error = CS_ERR_TRY_AGAIN;
 		goto error_put;
 	} else
 	if (res == -1) {
+		error = CS_ERR_LIBRARY;
 		goto error_put;
-	}
+	} else
 	if (res == 0) {
+		/* Means that the peer closed cleanly the socket. However, it should
+		 * happen only on BSD and Darwing systems since poll() returns a
+		 * POLLHUP event on other systems.
+		 */
+		error = CS_ERR_LIBRARY;
 		goto error_put;
 	}
 	ipc_instance->flow_control_state = 0;
-	if (buf == 1 || buf == 2) {
+	if (buf == MESSAGE_RES_OUTQ_NOT_EMPTY || buf == MESSAGE_RES_ENABLE_FLOWCONTROL) {
 		ipc_instance->flow_control_state = 1;
 	}
 	/*
@@ -799,10 +864,12 @@ coroipcc_dispatch_get (
 	 * This is just a notification of flow control starting at the addition
 	 * of a new pending message, not a message to dispatch
 	 */
-	if (buf == 2) {
+	if (buf == MESSAGE_RES_ENABLE_FLOWCONTROL) {
+		error = CS_ERR_TRY_AGAIN;
 		goto error_put;
 	}
-	if (buf == 3) {
+	if (buf == MESSAGE_RES_OUTQ_FLUSH_NR) {
+		error = CS_ERR_TRY_AGAIN;
 		goto error_put;
 	}
 
@@ -821,32 +888,44 @@ error_put:
 cs_error_t
 coroipcc_dispatch_put (hdb_handle_t handle)
 {
+#if _POSIX_THREAD_PROCESS_SHARED < 1
 	struct sembuf sop;
+#endif
 	coroipc_response_header_t *header;
 	struct ipc_instance *ipc_instance;
 	int res;
 	char *addr;
 	unsigned int read_idx;
 
-	res = hdb_error_to_cs (hdb_handle_get (&ipc_hdb, handle, (void **)&ipc_instance));
+	res = hdb_error_to_cs (hdb_handle_get_always (&ipc_hdb, handle, (void **)&ipc_instance));
 	if (res != CS_OK) {
 		return (res);
 	}
+#if _POSIX_THREAD_PROCESS_SHARED > 0
+retry_semwait:
+	res = sem_wait (&ipc_instance->control_buffer->sem2);
+	if (res == -1 && errno == EINTR) {
+		goto retry_semwait;
+	}
+#else
 	sop.sem_num = 2;
 	sop.sem_op = -1;
 	sop.sem_flg = 0;
 retry_semop:
 	res = semop (ipc_instance->semid, &sop, 1);
 	if (res == -1 && errno == EINTR) {
-		return (CS_ERR_TRY_AGAIN);
+		res = CS_ERR_TRY_AGAIN;
+		goto error_exit;
 	} else
 	if (res == -1 && errno == EACCES) {
 		priv_change_send (ipc_instance);
 		goto retry_semop;
 	} else
 	if (res == -1) {
-		return (CS_ERR_LIBRARY);
+		res = CS_ERR_LIBRARY;
+		goto error_exit;
 	}
+#endif
 
 	addr = ipc_instance->dispatch_buffer;
 
@@ -857,10 +936,15 @@ retry_semop:
 	/*
 	 * Put from dispatch get and also from this call's get
 	 */
+	res = CS_OK;
+	
+#if _POSIX_THREAD_PROCESS_SHARED < 1
+error_exit:
+#endif
 	hdb_handle_put (&ipc_hdb, handle);
 	hdb_handle_put (&ipc_hdb, handle);
 
-	return (CS_OK);
+	return (res);
 }
 
 cs_error_t
@@ -964,7 +1048,7 @@ coroipcc_zcb_alloc (
 		return (res);
 	}
 	map_size = size + header_size + sizeof (struct coroipcs_zc_header);
-	res = memory_map (path, "corosync_zerocopy-XXXXXX", &buf, size);
+	res = memory_map (path, "corosync_zerocopy-XXXXXX", &buf, map_size);
 	assert (res != -1);
 
 	req_coroipcc_zc_alloc.header.size = sizeof (mar_req_coroipcc_zc_alloc_t);
