@@ -47,27 +47,28 @@
 #include <stdio.h>
 #include <limits.h>
 #include <errno.h>
-#include <signal.h>
 #include <string.h>
 
 #include <corosync/corotypes.h>
+#include <corosync/coroipc_types.h>
 #include <corosync/cfg.h>
 #include <corosync/list.h>
-#include <corosync/queue.h>
 #include <corosync/mar_gen.h>
-#include <corosync/ipc_gen.h>
 #include <corosync/totem/totemip.h>
+#include <corosync/totem/totem.h>
 #include <corosync/ipc_cfg.h>
 #include <corosync/lcr/lcr_comp.h>
 #include <corosync/engine/logsys.h>
 #include <corosync/engine/coroapi.h>
+#include <corosync/corodefs.h>
 
-LOGSYS_DECLARE_SUBSYS ("CFG", LOG_INFO);
+LOGSYS_DECLARE_SUBSYS ("CFG");
 
 enum cfg_message_req_types {
         MESSAGE_REQ_EXEC_CFG_RINGREENABLE = 0,
 	MESSAGE_REQ_EXEC_CFG_KILLNODE = 1,
-	MESSAGE_REQ_EXEC_CFG_SHUTDOWN = 2
+	MESSAGE_REQ_EXEC_CFG_SHUTDOWN = 2,
+	MESSAGE_REQ_EXEC_CFG_CRYPTO_SET = 3
 };
 
 #define DEFAULT_SHUTDOWN_TIMEOUT 5
@@ -94,10 +95,10 @@ struct cfg_info
 
 static void cfg_confchg_fn (
 	enum totem_configuration_type configuration_type,
-	unsigned int *member_list, int member_list_entries,
-	unsigned int *left_list, int left_list_entries,
-	unsigned int *joined_list, int joined_list_entries,
-	struct memb_ring_id *ring_id);
+	const unsigned int *member_list, size_t member_list_entries,
+	const unsigned int *left_list, size_t left_list_entries,
+	const unsigned int *joined_list, size_t joined_list_entries,
+	const struct memb_ring_id *ring_id);
 
 static int cfg_exec_init_fn (struct corosync_api_v1 *corosync_api_v1);
 
@@ -108,70 +109,78 @@ static int cfg_lib_init_fn (void *conn);
 static int cfg_lib_exit_fn (void *conn);
 
 static void message_handler_req_exec_cfg_ringreenable (
-        void *message,
+        const void *message,
         unsigned int nodeid);
 
 static void message_handler_req_exec_cfg_killnode (
-        void *message,
+        const void *message,
         unsigned int nodeid);
 
 static void message_handler_req_exec_cfg_shutdown (
-        void *message,
+        const void *message,
+        unsigned int nodeid);
+
+static void message_handler_req_exec_cfg_crypto_set (
+        const void *message,
         unsigned int nodeid);
 
 static void exec_cfg_killnode_endian_convert (void *msg);
 
 static void message_handler_req_lib_cfg_ringstatusget (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_ringreenable (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_statetrack (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_statetrackstop (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_administrativestateset (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_administrativestateget (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_serviceload (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_serviceunload (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_killnode (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_tryshutdown (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_replytoshutdown (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_get_node_addrs (
 	void *conn,
-	void *msg);
+	const void *msg);
 
 static void message_handler_req_lib_cfg_local_get (
 	void *conn,
-	void *msg);
+	const void *msg);
+
+static void message_handler_req_lib_cfg_crypto_set (
+	void *conn,
+	const void *msg);
 
 /*
  * Service Handler Definition
@@ -180,80 +189,58 @@ static struct corosync_lib_handler cfg_lib_engine[] =
 {
 	{ /* 0 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_ringstatusget,
-		.response_size		= sizeof (struct res_lib_cfg_ringstatusget),
-		.response_id		= MESSAGE_RES_CFG_RINGSTATUSGET,
 		.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 1 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_ringreenable,
-		.response_size		= sizeof (struct res_lib_cfg_ringreenable),
-		.response_id		= MESSAGE_RES_CFG_RINGREENABLE,
 		.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 2 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_statetrack,
-		.response_size		= sizeof (struct res_lib_cfg_statetrack),
-		.response_id		= MESSAGE_RES_CFG_STATETRACKSTART,
 		.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 3 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_statetrackstop,
-		.response_size		= sizeof (struct res_lib_cfg_statetrackstop),
-		.response_id		= MESSAGE_RES_CFG_STATETRACKSTOP,
 		.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 4 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_administrativestateset,
-		.response_size		= sizeof (struct res_lib_cfg_administrativestateset),
-		.response_id		= MESSAGE_RES_CFG_ADMINISTRATIVESTATESET,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 5 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_administrativestateget,
-		.response_size		= sizeof (struct res_lib_cfg_administrativestateget),
-		.response_id		= MESSAGE_RES_CFG_ADMINISTRATIVESTATEGET,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 6 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_serviceload,
-		.response_size		= sizeof (struct res_lib_cfg_serviceload),
-		.response_id		= MESSAGE_RES_CFG_SERVICELOAD,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 7 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_serviceunload,
-		.response_size		= sizeof (struct res_lib_cfg_serviceunload),
-		.response_id		= MESSAGE_RES_CFG_SERVICEUNLOAD,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 8 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_killnode,
-		.response_size		= sizeof (struct res_lib_cfg_killnode),
-		.response_id		= MESSAGE_RES_CFG_KILLNODE,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 9 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_tryshutdown,
-		.response_size		= sizeof (struct res_lib_cfg_tryshutdown),
-		.response_id		= MESSAGE_RES_CFG_TRYSHUTDOWN,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 10 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_replytoshutdown,
-		.response_size		= sizeof (struct res_lib_cfg_replytoshutdown),
-		.response_id		= MESSAGE_RES_CFG_REPLYTOSHUTDOWN,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 11 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_get_node_addrs,
-		.response_size		= sizeof (struct res_lib_cfg_get_node_addrs),
-		.response_id		= MESSAGE_RES_CFG_GET_NODE_ADDRS,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 12 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_local_get,
-		.response_size		= sizeof (struct res_lib_cfg_local_get),
-		.response_id		= MESSAGE_RES_CFG_LOCAL_GET,
+		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
+	},
+	{ /* 13 */
+		.lib_handler_fn		= message_handler_req_lib_cfg_crypto_set,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	}
 };
@@ -269,6 +256,9 @@ static struct corosync_exec_handler cfg_exec_engine[] =
 	},
 	{ /* 2 */
 		.exec_handler_fn = message_handler_req_exec_cfg_shutdown,
+	},
+	{ /* 3 */
+		.exec_handler_fn = message_handler_req_exec_cfg_crypto_set,
 	}
 };
 
@@ -278,6 +268,7 @@ static struct corosync_exec_handler cfg_exec_engine[] =
 struct corosync_service_engine cfg_service_engine = {
 	.name					= "corosync configuration service",
 	.id					= CFG_SERVICE,
+	.priority				= 1,
 	.private_data_size			= sizeof(struct cfg_info),
 	.flow_control				= CS_LIB_FLOW_CONTROL_NOT_REQUIRED,
 	.allow_inquorate			= CS_LIB_ALLOW_INQUORATE,
@@ -289,6 +280,7 @@ struct corosync_service_engine cfg_service_engine = {
 	.exec_engine				= cfg_exec_engine,
 	.exec_engine_count			= 0, /* sizeof (cfg_aisexec_handler_fns) / sizeof (coroync_exec_handler), */
 	.confchg_fn				= cfg_confchg_fn,
+	.sync_mode				= CS_SYNC_V1
 };
 
 /*
@@ -324,25 +316,36 @@ static struct corosync_service_engine *cfg_get_service_engine_ver0 (void)
 	return (&cfg_service_engine);
 }
 
-__attribute__ ((constructor)) static void register_this_component (void) {
+#ifdef COROSYNC_SOLARIS
+void corosync_lcr_component_register (void);
+
+void corosync_lcr_component_register (void) {
+#else
+__attribute__ ((constructor)) static void corosync_lcr_component_register (void) {
+#endif
 	lcr_interfaces_set (&corosync_cfg_ver0[0], &cfg_service_engine_iface);
 
 	lcr_component_register (&cfg_comp_ver0);
 }
 
 struct req_exec_cfg_ringreenable {
-	mar_req_header_t header __attribute__((aligned(8)));
+	coroipc_request_header_t header __attribute__((aligned(8)));
         mar_message_source_t source __attribute__((aligned(8)));
 };
 
 struct req_exec_cfg_killnode {
-	mar_req_header_t header __attribute__((aligned(8)));
+	coroipc_request_header_t header __attribute__((aligned(8)));
         mar_uint32_t nodeid __attribute__((aligned(8)));
 	mar_name_t reason __attribute__((aligned(8)));
 };
 
+struct req_exec_cfg_crypto_set {
+	coroipc_request_header_t header __attribute__((aligned(8)));
+	mar_uint32_t type __attribute__((aligned(8)));
+};
+
 struct req_exec_cfg_shutdown {
-	mar_req_header_t header __attribute__((aligned(8)));
+	coroipc_request_header_t header __attribute__((aligned(8)));
 };
 
 /* IMPL */
@@ -350,6 +353,10 @@ struct req_exec_cfg_shutdown {
 static int cfg_exec_init_fn (
 	struct corosync_api_v1 *corosync_api_v1)
 {
+#ifdef COROSYNC_SOLARIS
+	logsys_subsys_init();
+#endif
+
 	api = corosync_api_v1;
 
 	list_init(&trackers_list);
@@ -358,17 +365,17 @@ static int cfg_exec_init_fn (
 
 static void cfg_confchg_fn (
 	enum totem_configuration_type configuration_type,
-	unsigned int *member_list, int member_list_entries,
-	unsigned int *left_list, int left_list_entries,
-	unsigned int *joined_list, int joined_list_entries,
-	struct memb_ring_id *ring_id)
+	const unsigned int *member_list, size_t member_list_entries,
+	const unsigned int *left_list, size_t left_list_entries,
+	const unsigned int *joined_list, size_t joined_list_entries,
+	const struct memb_ring_id *ring_id)
 {
 }
 
 /*
  * Tell other nodes we are shutting down
  */
-static int send_shutdown()
+static int send_shutdown(void)
 {
 	struct req_exec_cfg_shutdown req_exec_cfg_shutdown;
 	struct iovec iovec;
@@ -417,7 +424,7 @@ static void send_test_shutdown(void *only_conn, void *exclude_conn, int status)
 	LEAVE();
 }
 
-static void check_shutdown_status()
+static void check_shutdown_status(void)
 {
 	ENTER();
 
@@ -473,7 +480,7 @@ static void check_shutdown_status()
 			shutdown_con = NULL;
 		}
 
-		log_printf(LOG_DEBUG, "shutdown decision is: (yes count: %d, no count: %d) flags=%x\n", shutdown_yes, shutdown_no, shutdown_flags);
+		log_printf(LOGSYS_LEVEL_DEBUG, "shutdown decision is: (yes count: %d, no count: %d) flags=%x\n", shutdown_yes, shutdown_no, shutdown_flags);
 	}
 	LEAVE();
 }
@@ -563,11 +570,11 @@ static int cfg_lib_init_fn (void *conn)
  * Executive message handlers
  */
 static void message_handler_req_exec_cfg_ringreenable (
-        void *message,
+        const void *message,
         unsigned int nodeid)
 {
-	struct req_exec_cfg_ringreenable *req_exec_cfg_ringreenable =
-		(struct req_exec_cfg_ringreenable *)message;
+	const struct req_exec_cfg_ringreenable *req_exec_cfg_ringreenable
+	  = message;
 	struct res_lib_cfg_ringreenable res_lib_cfg_ringreenable;
 
 	ENTER();
@@ -596,18 +603,17 @@ static void exec_cfg_killnode_endian_convert (void *msg)
 
 
 static void message_handler_req_exec_cfg_killnode (
-        void *message,
+        const void *message,
         unsigned int nodeid)
 {
-	struct req_exec_cfg_killnode *req_exec_cfg_killnode =
-		(struct req_exec_cfg_killnode *)message;
+	const struct req_exec_cfg_killnode *req_exec_cfg_killnode = message;
 	cs_name_t reason;
 
 	ENTER();
-	log_printf(LOG_DEBUG, "request to kill node %d(us=%d): %s\n",  req_exec_cfg_killnode->nodeid, api->totem_nodeid_get(), reason.value);
+	log_printf(LOGSYS_LEVEL_DEBUG, "request to kill node %d(us=%d): %s\n",  req_exec_cfg_killnode->nodeid, api->totem_nodeid_get(), reason.value);
         if (req_exec_cfg_killnode->nodeid == api->totem_nodeid_get()) {
 		marshall_from_mar_name_t(&reason, &req_exec_cfg_killnode->reason);
-		log_printf(LOG_NOTICE, "Killed by node %d: %s\n",
+		log_printf(LOGSYS_LEVEL_NOTICE, "Killed by node %d: %s\n",
 			   nodeid, reason.value);
 		corosync_fatal_error(COROSYNC_FATAL_ERROR_EXIT);
 	}
@@ -618,31 +624,44 @@ static void message_handler_req_exec_cfg_killnode (
  * Self shutdown
  */
 static void message_handler_req_exec_cfg_shutdown (
-        void *message,
+        const void *message,
         unsigned int nodeid)
 {
 	ENTER();
 
-	log_printf(LOG_NOTICE, "Node %d was shut down by sysadmin\n", nodeid);
+	log_printf(LOGSYS_LEVEL_NOTICE, "Node %d was shut down by sysadmin\n", nodeid);
 	if (nodeid == api->totem_nodeid_get()) {
-		corosync_fatal_error(COROSYNC_FATAL_ERROR_EXIT);
+		api->shutdown_request();
 	}
 	LEAVE();
 }
 
+static void message_handler_req_exec_cfg_crypto_set (
+        const void *message,
+        unsigned int nodeid)
+{
+	const struct req_exec_cfg_crypto_set *req_exec_cfg_crypto_set = message;
+	ENTER();
+
+	log_printf(LOGSYS_LEVEL_NOTICE, "Node %d requested set crypto to %d\n", nodeid, req_exec_cfg_crypto_set->type);
+
+	api->totem_crypto_set(req_exec_cfg_crypto_set->type);
+	LEAVE();
+}
+
 
 /*
  * Library Interface Implementation
  */
 static void message_handler_req_lib_cfg_ringstatusget (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
 	struct res_lib_cfg_ringstatusget res_lib_cfg_ringstatusget;
 	struct totem_ip_address interfaces[INTERFACE_MAX];
 	unsigned int iface_count;
 	char **status;
-	char *totem_ip_string;
+	const char *totem_ip_string;
 	unsigned int i;
 
 	ENTER();
@@ -660,7 +679,8 @@ static void message_handler_req_lib_cfg_ringstatusget (
 	res_lib_cfg_ringstatusget.interface_count = iface_count;
 
 	for (i = 0; i < iface_count; i++) {
-		totem_ip_string = (char *)api->totem_ip_print (&interfaces[i]);
+		totem_ip_string
+		  = (const char *)api->totem_ip_print (&interfaces[i]);
 		strcpy ((char *)&res_lib_cfg_ringstatusget.interface_status[i],
 			status[i]);
 		strcpy ((char *)&res_lib_cfg_ringstatusget.interface_name[i],
@@ -676,7 +696,7 @@ static void message_handler_req_lib_cfg_ringstatusget (
 
 static void message_handler_req_lib_cfg_ringreenable (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
 	struct req_exec_cfg_ringreenable req_exec_cfg_ringreenable;
 	struct iovec iovec;
@@ -698,7 +718,7 @@ static void message_handler_req_lib_cfg_ringreenable (
 
 static void message_handler_req_lib_cfg_statetrack (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
 	struct cfg_info *ci = (struct cfg_info *)api->ipc_private_data_get (conn);
 	struct res_lib_cfg_statetrack res_lib_cfg_statetrack;
@@ -734,7 +754,7 @@ static void message_handler_req_lib_cfg_statetrack (
 
 static void message_handler_req_lib_cfg_statetrackstop (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
 	struct cfg_info *ci = (struct cfg_info *)api->ipc_private_data_get (conn);
 //	struct req_lib_cfg_statetrackstop *req_lib_cfg_statetrackstop = (struct req_lib_cfg_statetrackstop *)message;
@@ -746,7 +766,7 @@ static void message_handler_req_lib_cfg_statetrackstop (
 
 static void message_handler_req_lib_cfg_administrativestateset (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
 //	struct req_lib_cfg_administrativestateset *req_lib_cfg_administrativestateset = (struct req_lib_cfg_administrativestateset *)message;
 
@@ -755,7 +775,7 @@ static void message_handler_req_lib_cfg_administrativestateset (
 }
 static void message_handler_req_lib_cfg_administrativestateget (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
 //	struct req_lib_cfg_administrativestateget *req_lib_cfg_administrativestateget = (struct req_lib_cfg_administrativestateget *)message;
 	ENTER();
@@ -764,16 +784,15 @@ static void message_handler_req_lib_cfg_administrativestateget (
 
 static void message_handler_req_lib_cfg_serviceload (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
-	struct req_lib_cfg_serviceload *req_lib_cfg_serviceload =
-		(struct req_lib_cfg_serviceload *)msg;
+	const struct req_lib_cfg_serviceload *req_lib_cfg_serviceload = msg;
 	struct res_lib_cfg_serviceload res_lib_cfg_serviceload;
 
 	ENTER();
 	api->service_link_and_init (
 		api,
-		(char *)req_lib_cfg_serviceload->service_name,
+		(const char *)req_lib_cfg_serviceload->service_name,
 		req_lib_cfg_serviceload->service_ver);
 
 	res_lib_cfg_serviceload.header.id = MESSAGE_RES_CFG_SERVICEUNLOAD;
@@ -788,16 +807,15 @@ static void message_handler_req_lib_cfg_serviceload (
 
 static void message_handler_req_lib_cfg_serviceunload (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
-	struct req_lib_cfg_serviceunload *req_lib_cfg_serviceunload =
-		(struct req_lib_cfg_serviceunload *)msg;
+	const struct req_lib_cfg_serviceunload *req_lib_cfg_serviceunload = msg;
 	struct res_lib_cfg_serviceunload res_lib_cfg_serviceunload;
 
 	ENTER();
 	api->service_unlink_and_exit (
 		api,
-		(char *)req_lib_cfg_serviceunload->service_name,
+		(const char *)req_lib_cfg_serviceunload->service_name,
 		req_lib_cfg_serviceunload->service_ver);
 	res_lib_cfg_serviceunload.header.id = MESSAGE_RES_CFG_SERVICEUNLOAD;
 	res_lib_cfg_serviceunload.header.size = sizeof (struct res_lib_cfg_serviceunload);
@@ -812,9 +830,9 @@ static void message_handler_req_lib_cfg_serviceunload (
 
 static void message_handler_req_lib_cfg_killnode (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
-	struct req_lib_cfg_killnode *req_lib_cfg_killnode = (struct req_lib_cfg_killnode *)msg;
+	const struct req_lib_cfg_killnode *req_lib_cfg_killnode = msg;
 	struct res_lib_cfg_killnode res_lib_cfg_killnode;
 	struct req_exec_cfg_killnode req_exec_cfg_killnode;
 	struct iovec iovec;
@@ -846,16 +864,16 @@ static void message_handler_req_lib_cfg_killnode (
 
 static void message_handler_req_lib_cfg_tryshutdown (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
 	struct cfg_info *ci = (struct cfg_info *)api->ipc_private_data_get (conn);
-	struct req_lib_cfg_tryshutdown *req_lib_cfg_tryshutdown = (struct req_lib_cfg_tryshutdown *)msg;
-	struct res_lib_cfg_tryshutdown res_lib_cfg_tryshutdown;
+	const struct req_lib_cfg_tryshutdown *req_lib_cfg_tryshutdown = msg;
 	struct list_head *iter;
 
 	ENTER();
 
 	if (req_lib_cfg_tryshutdown->flags == CFG_SHUTDOWN_FLAG_IMMEDIATE) {
+		struct res_lib_cfg_tryshutdown res_lib_cfg_tryshutdown;
 
 		/*
 		 * Tell other nodes
@@ -979,10 +997,10 @@ static void message_handler_req_lib_cfg_tryshutdown (
 
 static void message_handler_req_lib_cfg_replytoshutdown (
 	void *conn,
-	void *msg)
+	const void *msg)
 {
 	struct cfg_info *ci = (struct cfg_info *)api->ipc_private_data_get (conn);
-	struct req_lib_cfg_replytoshutdown *req_lib_cfg_replytoshutdown = (struct req_lib_cfg_replytoshutdown *)msg;
+	const struct req_lib_cfg_replytoshutdown *req_lib_cfg_replytoshutdown = msg;
 	struct res_lib_cfg_replytoshutdown res_lib_cfg_replytoshutdown;
 	int status = CS_OK;
 
@@ -1013,7 +1031,8 @@ exit_fn:
 	LEAVE();
 }
 
-static void message_handler_req_lib_cfg_get_node_addrs (void *conn, void *msg)
+static void message_handler_req_lib_cfg_get_node_addrs (void *conn,
+							const void *msg)
 {
 	struct totem_ip_address node_ifs[INTERFACE_MAX];
 	char buf[PIPE_BUF];
@@ -1021,13 +1040,14 @@ static void message_handler_req_lib_cfg_get_node_addrs (void *conn, void *msg)
 	unsigned int num_interfaces = 0;
 	int ret = CS_OK;
 	int i;
-	struct req_lib_cfg_get_node_addrs *req_lib_cfg_get_node_addrs = (struct req_lib_cfg_get_node_addrs *)msg;
+	const struct req_lib_cfg_get_node_addrs *req_lib_cfg_get_node_addrs = msg;
 	struct res_lib_cfg_get_node_addrs *res_lib_cfg_get_node_addrs = (struct res_lib_cfg_get_node_addrs *)buf;
+	unsigned int nodeid = req_lib_cfg_get_node_addrs->nodeid;
 
-	if (req_lib_cfg_get_node_addrs->nodeid == 0)
-		req_lib_cfg_get_node_addrs->nodeid = api->totem_nodeid_get();
+	if (nodeid == 0)
+		nodeid = api->totem_nodeid_get();
 
-	api->totem_ifaces_get(req_lib_cfg_get_node_addrs->nodeid, node_ifs, &status, &num_interfaces);
+	api->totem_ifaces_get(nodeid, node_ifs, &status, &num_interfaces);
 
 	res_lib_cfg_get_node_addrs->header.size = sizeof(struct res_lib_cfg_get_node_addrs) + (num_interfaces * TOTEMIP_ADDRLEN);
 	res_lib_cfg_get_node_addrs->header.id = MESSAGE_RES_CFG_GET_NODE_ADDRS;
@@ -1045,7 +1065,7 @@ static void message_handler_req_lib_cfg_get_node_addrs (void *conn, void *msg)
 	api->ipc_response_send(conn, res_lib_cfg_get_node_addrs, res_lib_cfg_get_node_addrs->header.size);
 }
 
-static void message_handler_req_lib_cfg_local_get (void *conn, void *message)
+static void message_handler_req_lib_cfg_local_get (void *conn, const void *msg)
 {
 	struct res_lib_cfg_local_get res_lib_cfg_local_get;
 
@@ -1057,3 +1077,40 @@ static void message_handler_req_lib_cfg_local_get (void *conn, void *message)
 	api->ipc_response_send(conn, &res_lib_cfg_local_get,
 		sizeof(res_lib_cfg_local_get));
 }
+
+
+static void message_handler_req_lib_cfg_crypto_set (
+	void *conn,
+	const void *msg)
+{
+	const struct req_lib_cfg_crypto_set *req_lib_cfg_crypto_set = msg;
+	struct res_lib_cfg_crypto_set res_lib_cfg_crypto_set;
+	struct req_exec_cfg_crypto_set req_exec_cfg_crypto_set;
+	struct iovec iovec;
+	int ret = CS_ERR_INVALID_PARAM;
+
+	req_exec_cfg_crypto_set.header.size =
+		sizeof (struct req_exec_cfg_crypto_set);
+	req_exec_cfg_crypto_set.header.id = SERVICE_ID_MAKE (CFG_SERVICE,
+		MESSAGE_REQ_EXEC_CFG_CRYPTO_SET);
+
+	/*
+	 * Set it locally first so we can tell if it is allowed
+	 */
+	if (api->totem_crypto_set(req_lib_cfg_crypto_set->type) == 0) {
+
+		req_exec_cfg_crypto_set.type = req_lib_cfg_crypto_set->type;
+
+		iovec.iov_base = (char *)&req_exec_cfg_crypto_set;
+		iovec.iov_len = sizeof (struct req_exec_cfg_crypto_set);
+		assert (api->totem_mcast (&iovec, 1, TOTEM_SAFE) == 0);
+		ret = CS_OK;
+	}
+
+	res_lib_cfg_crypto_set.header.size = sizeof(res_lib_cfg_crypto_set);
+	res_lib_cfg_crypto_set.header.id = MESSAGE_RES_CFG_CRYPTO_SET;
+	res_lib_cfg_crypto_set.header.error = ret;
+
+	api->ipc_response_send(conn, &res_lib_cfg_crypto_set,
+		sizeof(res_lib_cfg_crypto_set));
+}
