@@ -7,7 +7,7 @@
  * Author: Steven Dake (sdake@redhat.com)
  *
  * This software licensed under BSD license, the text of which follows:
- * 
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *
@@ -46,21 +46,24 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
-#include <signal.h>
 #include <time.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
 #include <corosync/swab.h>
 #include <corosync/corotypes.h>
-#include <corosync/ipc_gen.h>
+#include <corosync/coroipc_types.h>
+#include <corosync/corodefs.h>
+#include <corosync/mar_gen.h>
 #include <corosync/lcr/lcr_comp.h>
 #include <corosync/engine/coroapi.h>
-#include <corosync/ipc_evs.h>
-#include <corosync/list.h>
 #include <corosync/engine/logsys.h>
+#include <corosync/list.h>
 
-LOGSYS_DECLARE_SUBSYS ("EVS", LOG_INFO);
+#include <corosync/evs.h>
+#include <corosync/ipc_evs.h>
+
+LOGSYS_DECLARE_SUBSYS ("EVS");
 
 enum evs_exec_message_req_types {
 	MESSAGE_REQ_EXEC_EVS_MCAST = 0
@@ -74,20 +77,20 @@ static int evs_exec_init_fn (
 
 static void evs_confchg_fn (
 	enum totem_configuration_type configuration_type,
-	unsigned int *member_list, int member_list_entries,
-	unsigned int *left_list, int left_list_entries,
-	unsigned int *joined_list, int joined_list_entries,
-	struct memb_ring_id *ring_id);
+	const unsigned int *member_list, size_t member_list_entries,
+	const unsigned int *left_list, size_t left_list_entries,
+	const unsigned int *joined_list, size_t joined_list_entries,
+	const struct memb_ring_id *ring_id);
 
-static void message_handler_req_exec_mcast (void *msg, unsigned int nodeid);
+static void message_handler_req_exec_mcast (const void *msg, unsigned int nodeid);
 
 static void req_exec_mcast_endian_convert (void *msg);
 
-static void message_handler_req_evs_join (void *conn, void *msg);
-static void message_handler_req_evs_leave (void *conn, void *msg);
-static void message_handler_req_evs_mcast_joined (void *conn, void *msg);
-static void message_handler_req_evs_mcast_groups (void *conn, void *msg);
-static void message_handler_req_evs_membership_get (void *conn, void *msg);
+static void message_handler_req_evs_join (void *conn, const void *msg);
+static void message_handler_req_evs_leave (void *conn, const void *msg);
+static void message_handler_req_evs_mcast_joined (void *conn, const void *msg);
+static void message_handler_req_evs_mcast_groups (void *conn, const void *msg);
+static void message_handler_req_evs_membership_get (void *conn, const void *msg);
 
 static int evs_lib_init_fn (void *conn);
 static int evs_lib_exit_fn (void *conn);
@@ -98,39 +101,29 @@ struct evs_pd {
 	struct list_head list;
 	void *conn;
 };
-	
+
 static struct corosync_api_v1 *api;
 
 static struct corosync_lib_handler evs_lib_engine[] =
 {
 	{ /* 0 */
 		.lib_handler_fn				= message_handler_req_evs_join,
-		.response_size				= sizeof (struct res_lib_evs_join),
-		.response_id				= MESSAGE_RES_EVS_JOIN,
 		.flow_control				= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 1 */
 		.lib_handler_fn				= message_handler_req_evs_leave,
-		.response_size				= sizeof (struct res_lib_evs_leave),
-		.response_id				= MESSAGE_RES_EVS_LEAVE,
 		.flow_control				= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	},
 	{ /* 2 */
 		.lib_handler_fn				= message_handler_req_evs_mcast_joined,
-		.response_size				= sizeof (struct res_lib_evs_mcast_joined),
-		.response_id				= MESSAGE_RES_EVS_MCAST_JOINED,
 		.flow_control				= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 3 */
 		.lib_handler_fn				= message_handler_req_evs_mcast_groups,
-		.response_size				= sizeof (struct res_lib_evs_mcast_groups),
-		.response_id				= MESSAGE_RES_EVS_MCAST_GROUPS,
 		.flow_control				= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 4 */
 		.lib_handler_fn				= message_handler_req_evs_membership_get,
-		.response_size				= sizeof (struct res_lib_evs_membership_get),
-		.response_id				= MESSAGE_RES_EVS_MEMBERSHIP_GET,
 		.flow_control				= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	}
 };
@@ -146,8 +139,9 @@ static struct corosync_exec_handler evs_exec_engine[] =
 struct corosync_service_engine evs_service_engine = {
 	.name			= "corosync extended virtual synchrony service",
 	.id			= EVS_SERVICE,
+	.priority		= 1,
 	.private_data_size	= sizeof (struct evs_pd),
-	.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED, 
+	.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED,
 	.lib_init_fn		= evs_lib_init_fn,
 	.lib_exit_fn		= evs_lib_exit_fn,
 	.lib_engine		= evs_lib_engine,
@@ -156,7 +150,8 @@ struct corosync_service_engine evs_service_engine = {
 	.exec_engine_count	= sizeof (evs_exec_engine) / sizeof (struct corosync_exec_handler),
 	.confchg_fn		= evs_confchg_fn,
 	.exec_init_fn		= evs_exec_init_fn,
-	.exec_dump_fn		= NULL
+	.exec_dump_fn		= NULL,
+	.sync_mode		= CS_SYNC_V1
 };
 
 static DECLARE_LIST_INIT (confchg_notify);
@@ -195,7 +190,13 @@ static struct corosync_service_engine *evs_get_service_engine_ver0 (void)
 	return (&evs_service_engine);
 }
 
-__attribute__ ((constructor)) static void evs_comp_register (void) {
+#ifdef COROSYNC_SOLARIS
+void corosync_lcr_component_register (void);
+
+void corosync_lcr_component_register (void) {
+#else
+__attribute__ ((constructor)) static void corosync_lcr_component_register (void) {
+#endif
 	lcr_interfaces_set (&corosync_evs_ver0[0], &evs_service_engine_iface);
 
 	lcr_component_register (&evs_comp_ver0);
@@ -204,6 +205,10 @@ __attribute__ ((constructor)) static void evs_comp_register (void) {
 static int evs_exec_init_fn (
 	struct corosync_api_v1 *corosync_api)
 {
+#ifdef COROSYNC_SOLARIS
+	logsys_subsys_init();
+#endif
+
 	api = corosync_api;
 
 	return 0;
@@ -213,10 +218,10 @@ struct res_evs_confchg_callback res_evs_confchg_callback;
 
 static void evs_confchg_fn (
 	enum totem_configuration_type configuration_type,
-	unsigned int *member_list, int member_list_entries,
-	unsigned int *left_list, int left_list_entries,
-	unsigned int *joined_list, int joined_list_entries,
-	struct memb_ring_id *ring_id)
+	const unsigned int *member_list, size_t member_list_entries,
+	const unsigned int *left_list, size_t left_list_entries,
+	const unsigned int *joined_list, size_t joined_list_entries,
+	const struct memb_ring_id *ring_id)
 {
 	struct list_head *list;
 	struct evs_pd *evs_pd;
@@ -255,7 +260,7 @@ static int evs_lib_init_fn (void *conn)
 {
 	struct evs_pd *evs_pd = (struct evs_pd *)api->ipc_private_data_get (conn);
 
-	log_printf (LOG_LEVEL_DEBUG, "Got request to initalize evs service.\n");
+	log_printf (LOGSYS_LEVEL_DEBUG, "Got request to initalize evs service.\n");
 
 	evs_pd->groups = NULL;
 	evs_pd->group_entries = 0;
@@ -263,7 +268,7 @@ static int evs_lib_init_fn (void *conn)
 	list_init (&evs_pd->list);
 	list_add (&evs_pd->list, &confchg_notify);
 
-	api->ipc_response_send (conn, &res_evs_confchg_callback,
+	api->ipc_dispatch_send (conn, &res_evs_confchg_callback,
 		sizeof (res_evs_confchg_callback));
 
 	return (0);
@@ -277,10 +282,10 @@ static int evs_lib_exit_fn (void *conn)
 	return (0);
 }
 
-static void message_handler_req_evs_join (void *conn, void *msg)
+static void message_handler_req_evs_join (void *conn, const void *msg)
 {
 	cs_error_t error = CS_OK;
-	struct req_lib_evs_join *req_lib_evs_join = (struct req_lib_evs_join *)msg;
+	const struct req_lib_evs_join *req_lib_evs_join = msg;
 	struct res_lib_evs_join res_lib_evs_join;
 	void *addr;
 	struct evs_pd *evs_pd = (struct evs_pd *)api->ipc_private_data_get (conn);
@@ -290,7 +295,7 @@ static void message_handler_req_evs_join (void *conn, void *msg)
 		goto exit_error;
 	}
 
-	addr = realloc (evs_pd->groups, sizeof (struct evs_group) * 
+	addr = realloc (evs_pd->groups, sizeof (struct evs_group) *
 		(evs_pd->group_entries + req_lib_evs_join->group_entries));
 	if (addr == NULL) {
 		error = CS_ERR_NO_MEMORY;
@@ -313,9 +318,9 @@ exit_error:
 		sizeof (struct res_lib_evs_join));
 }
 
-static void message_handler_req_evs_leave (void *conn, void *msg)
+static void message_handler_req_evs_leave (void *conn, const void *msg)
 {
-	struct req_lib_evs_leave *req_lib_evs_leave = (struct req_lib_evs_leave *)msg;
+	const struct req_lib_evs_leave *req_lib_evs_leave = msg;
 	struct res_lib_evs_leave res_lib_evs_leave;
 	cs_error_t error = CS_OK;
 	int error_index;
@@ -359,10 +364,10 @@ static void message_handler_req_evs_leave (void *conn, void *msg)
 		sizeof (struct res_lib_evs_leave));
 }
 
-static void message_handler_req_evs_mcast_joined (void *conn, void *msg)
+static void message_handler_req_evs_mcast_joined (void *conn, const void *msg)
 {
 	cs_error_t error = CS_ERR_TRY_AGAIN;
-	struct req_lib_evs_mcast_joined *req_lib_evs_mcast_joined = (struct req_lib_evs_mcast_joined *)msg;
+	const struct req_lib_evs_mcast_joined *req_lib_evs_mcast_joined = msg;
 	struct res_lib_evs_mcast_joined res_lib_evs_mcast_joined;
 	struct iovec req_exec_evs_mcast_iovec[3];
 	struct req_exec_evs_mcast req_exec_evs_mcast;
@@ -399,14 +404,14 @@ static void message_handler_req_evs_mcast_joined (void *conn, void *msg)
 		sizeof (struct res_lib_evs_mcast_joined));
 }
 
-static void message_handler_req_evs_mcast_groups (void *conn, void *msg)
+static void message_handler_req_evs_mcast_groups (void *conn, const void *msg)
 {
 	cs_error_t error = CS_ERR_TRY_AGAIN;
-	struct req_lib_evs_mcast_groups *req_lib_evs_mcast_groups = (struct req_lib_evs_mcast_groups *)msg;
+	const struct req_lib_evs_mcast_groups *req_lib_evs_mcast_groups = msg;
 	struct res_lib_evs_mcast_groups res_lib_evs_mcast_groups;
 	struct iovec req_exec_evs_mcast_iovec[3];
 	struct req_exec_evs_mcast req_exec_evs_mcast;
-	char *msg_addr;
+	const char *msg_addr;
 	int res;
 
 	req_exec_evs_mcast.header.size = sizeof (struct req_exec_evs_mcast) +
@@ -418,17 +423,17 @@ static void message_handler_req_evs_mcast_groups (void *conn, void *msg)
 	req_exec_evs_mcast.msg_len = req_lib_evs_mcast_groups->msg_len;
 	req_exec_evs_mcast.group_entries = req_lib_evs_mcast_groups->group_entries;
 
-	msg_addr = (char *)req_lib_evs_mcast_groups +
-		sizeof (struct req_lib_evs_mcast_groups) + 
+	msg_addr = (const char *)req_lib_evs_mcast_groups +
+		sizeof (struct req_lib_evs_mcast_groups) +
 		(sizeof (struct evs_group) * req_lib_evs_mcast_groups->group_entries);
 
 	req_exec_evs_mcast_iovec[0].iov_base = (char *)&req_exec_evs_mcast;
 	req_exec_evs_mcast_iovec[0].iov_len = sizeof (req_exec_evs_mcast);
 	req_exec_evs_mcast_iovec[1].iov_base = (char *)&req_lib_evs_mcast_groups->groups;
 	req_exec_evs_mcast_iovec[1].iov_len = sizeof (struct evs_group) * req_lib_evs_mcast_groups->group_entries;
-	req_exec_evs_mcast_iovec[2].iov_base = msg_addr;
+	req_exec_evs_mcast_iovec[2].iov_base = (void *) msg_addr; /* discard const */
 	req_exec_evs_mcast_iovec[2].iov_len = req_lib_evs_mcast_groups->msg_len;
-	
+
 	res = api->totem_mcast (req_exec_evs_mcast_iovec, 3, TOTEM_AGREED);
 	if (res == 0) {
 		error = CS_OK;
@@ -442,7 +447,7 @@ static void message_handler_req_evs_mcast_groups (void *conn, void *msg)
 		sizeof (struct res_lib_evs_mcast_groups));
 }
 
-static void message_handler_req_evs_membership_get (void *conn, void *msg)
+static void message_handler_req_evs_membership_get (void *conn, const void *msg)
 {
 	struct res_lib_evs_membership_get res_lib_evs_membership_get;
 
@@ -471,12 +476,12 @@ static void req_exec_mcast_endian_convert (void *msg)
 }
 
 static void message_handler_req_exec_mcast (
-	void *msg,
+	const void *msg,
 	unsigned int nodeid)
 {
-	struct req_exec_evs_mcast *req_exec_evs_mcast = (struct req_exec_evs_mcast *)msg;
+	const struct req_exec_evs_mcast *req_exec_evs_mcast = msg;
 	struct res_evs_deliver_callback res_evs_deliver_callback;
-	char *msg_addr;
+	const char *msg_addr;
 	struct list_head *list;
 	int found = 0;
 	int i, j;
@@ -489,7 +494,7 @@ static void message_handler_req_exec_mcast (
 	res_evs_deliver_callback.header.error = CS_OK;
 	res_evs_deliver_callback.msglen = req_exec_evs_mcast->msg_len;
 
-	msg_addr = (char *)req_exec_evs_mcast + sizeof (struct req_exec_evs_mcast) + 
+	msg_addr = (const char *)req_exec_evs_mcast + sizeof (struct req_exec_evs_mcast) +
 		(sizeof (struct evs_group) * req_exec_evs_mcast->group_entries);
 
 	for (list = confchg_notify.next; list != &confchg_notify; list = list->next) {
@@ -513,9 +518,9 @@ static void message_handler_req_exec_mcast (
 
 		if (found) {
 			res_evs_deliver_callback.local_nodeid = nodeid;
-			iov[0].iov_base = &res_evs_deliver_callback;
+			iov[0].iov_base = (void *)&res_evs_deliver_callback;
 			iov[0].iov_len = sizeof (struct res_evs_deliver_callback);
-			iov[1].iov_base = msg_addr;
+			iov[1].iov_base = (void *) msg_addr; /* discard const */
 			iov[1].iov_len = req_exec_evs_mcast->msg_len;
 
 			api->ipc_dispatch_iov_send (
@@ -525,4 +530,3 @@ static void message_handler_req_exec_mcast (
 		}
 	}
 }
-
