@@ -38,9 +38,9 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
-#include <signal.h>
 #include <unistd.h>
 #include <string.h>
+#include <ctype.h>
 #include <sys/types.h>
 #include <sys/un.h>
 
@@ -51,6 +51,7 @@
 #define SEPERATOR_STR "."
 #define OBJ_NAME_SIZE 512
 
+
 typedef enum {
 	ACTION_READ,
 	ACTION_WRITE,
@@ -71,65 +72,106 @@ static void tail_key_changed(confdb_handle_t handle,
 	confdb_change_type_t change_type,
 	hdb_handle_t parent_object_handle,
 	hdb_handle_t object_handle,
-	void *object_name,
-	int  object_name_len,
-	void *key_name,
-	int key_name_len,
-	void *key_value,
-	int key_value_len);
+	const void *object_name,
+	size_t  object_name_len,
+	const void *key_name,
+	size_t key_name_len,
+	const void *key_value,
+	size_t key_value_len);
 
 static void tail_object_created(confdb_handle_t handle,
 	hdb_handle_t parent_object_handle,
 	hdb_handle_t object_handle,
-	uint8_t *name_pt,
-	int  name_len);
+	const void *name_pt,
+	size_t name_len);
 
 static void tail_object_deleted(confdb_handle_t handle,
 	hdb_handle_t parent_object_handle,
-	uint8_t *name_pt,
-	int  name_len);
+	const void *name_pt,
+	size_t name_len);
 
-confdb_callbacks_t callbacks = {
+static void create_object(confdb_handle_t handle, char * name_pt);
+static void write_key(confdb_handle_t handle, char * path_pt);
+static void get_parent_name(const char * name_pt, char * parent_name);
+
+static confdb_callbacks_t callbacks = {
 	.confdb_key_change_notify_fn = tail_key_changed,
 	.confdb_object_create_change_notify_fn = tail_object_created,
 	.confdb_object_delete_change_notify_fn = tail_object_deleted,
 };
 
+static int debug = 0;
 static int action;
 
+static void print_key (char *key_name, void *value, size_t value_len, confdb_value_types_t type)
+{
+	switch (type) {
+		case CONFDB_VALUETYPE_INT16:
+			printf ("%s=%hd\n", key_name,
+					  *(int16_t*)value);
+			break;
+		case CONFDB_VALUETYPE_UINT16:
+			printf ("%s=%hu\n", key_name,
+					  *(uint16_t*)value);
+			break;
+		case CONFDB_VALUETYPE_INT32:
+			printf ("%s=%d\n", key_name,
+					  *(int32_t*)value);
+			break;
+		case CONFDB_VALUETYPE_UINT32:
+			printf ("%s=%u\n", key_name,
+					  *(uint32_t*)value);
+			break;
+		case CONFDB_VALUETYPE_INT64:
+			printf ("%s=%lld\n", key_name,
+					  *(int64_t*)value);
+			break;
+		case CONFDB_VALUETYPE_UINT64:
+			printf ("%s=%llu\n", key_name,
+					  *(uint64_t*)value);
+			break;
+		case CONFDB_VALUETYPE_STRING:
+			printf ("%s=%s\n", key_name, (char*)value);
+			break;
+		default:
+		case CONFDB_VALUETYPE_ANY:
+			printf ("%s=**binary**(%d)\n", key_name, type);
+			break;
+	}
+}
+
 /* Recursively dump the object tree */
 static void print_config_tree(confdb_handle_t handle, hdb_handle_t parent_object_handle, char * parent_name)
 {
 	hdb_handle_t object_handle;
 	char object_name[OBJ_NAME_SIZE];
-	int object_name_len;
+	size_t object_name_len;
 	char key_name[OBJ_NAME_SIZE];
-	int key_name_len;
 	char key_value[OBJ_NAME_SIZE];
-	int key_value_len;
+	size_t key_value_len;
 	cs_error_t res;
 	int children_printed;
+	confdb_value_types_t type;
 
 	/* Show the keys */
 	res = confdb_key_iter_start(handle, parent_object_handle);
 	if (res != CS_OK) {
-		fprintf(stderr, "error resetting key iterator for object %llu: %d\n", parent_object_handle, res);
+		fprintf(stderr, "error resetting key iterator for object "HDB_X_FORMAT" %d\n", parent_object_handle, res);
 		exit(EXIT_FAILURE);
 	}
 	children_printed = 0;
 
-	while ( (res = confdb_key_iter(handle,
+	while ( (res = confdb_key_iter_typed(handle,
 								   parent_object_handle,
 								   key_name,
-								   &key_name_len,
 								   key_value,
-								   &key_value_len)) == CS_OK) {
-		key_name[key_name_len] = '\0';
+								   &key_value_len,
+								   &type)) == CS_OK) {
 		key_value[key_value_len] = '\0';
 		if (parent_name != NULL)
-			printf("%s%c%s=%s\n", parent_name, SEPERATOR,key_name, key_value);
-		else
-			printf("%s=%s\n", key_name, key_value);
+			printf("%s%c", parent_name, SEPERATOR);
+
+		print_key(key_name, key_value, key_value_len, type);
 
 		children_printed++;
 	}
@@ -137,7 +179,7 @@ static void print_config_tree(confdb_handle_t handle, hdb_handle_t parent_object
 	/* Show sub-objects */
 	res = confdb_object_iter_start(handle, parent_object_handle);
 	if (res != CS_OK) {
-		fprintf(stderr, "error resetting object iterator for object %llu: %d\n", parent_object_handle, res);
+		fprintf(stderr, "error resetting object iterator for object "HDB_X_FORMAT" %d\n", parent_object_handle, res);
 		exit(EXIT_FAILURE);
 	}
 
@@ -162,6 +204,82 @@ static void print_config_tree(confdb_handle_t handle, hdb_handle_t parent_object
 	}
 }
 
+static int read_in_config_file (char * filename)
+{
+	confdb_handle_t handle;
+	int result;
+	int ignore;
+	int c;
+	FILE* fh;
+	char buf[1024];
+	char * line;
+	char * end;
+	char parent_name[OBJ_NAME_SIZE];
+
+	if (access (filename, R_OK) != 0) {
+		perror ("Couldn't access file.");
+		return -1;
+	}
+
+	fh = fopen(filename, "r");
+	if (fh == NULL) {
+		perror ("Couldn't open file.");
+		return -1;
+	}
+	result = confdb_initialize (&handle, &callbacks);
+	if (result != CONFDB_OK) {
+		fprintf (stderr, "Could not initialize objdb library. Error %d\n", result);
+		fclose (fh);
+		return -1;
+	}
+
+	while (fgets (buf, 1024, fh) != NULL) {
+		/* find the first real character, if it is
+		 * a '#' then ignore this line.
+		 * else process.
+		 * if no real characters then also ignore.
+		 */
+		ignore = 1;
+		for (c = 0; c < 1024; c++) {
+			if (isblank (buf[c]))
+				continue;
+
+			if (buf[c] == '#' || buf[c] == '\n') {
+				ignore = 1;
+				break;
+			}
+			ignore = 0;
+			line = &buf[c];
+			break;
+		}
+		if (ignore == 1)
+			continue;
+
+		/* kill the \n */
+		end = strchr (line, '\n');
+		if (end != NULL)
+			*end = '\0';
+
+		if (debug == 2)
+			printf ("%d: %s\n", __LINE__, line);
+
+		/* find the parent object */
+		get_parent_name(line, parent_name);
+
+		if (debug == 2)
+			printf ("%d: %s\n", __LINE__, parent_name);
+
+		/* create the object */
+		create_object (handle, parent_name);
+		/* write the attribute */
+		write_key (handle, line);
+	}
+
+	confdb_finalize (handle);
+	fclose (fh);
+	return 0;
+}
+
 static int print_all(void)
 {
 	confdb_handle_t handle;
@@ -190,6 +308,7 @@ static int print_help(void)
 	printf ("        corosync-objctl -w object%cchild_obj.key=value ... Create a key\n", SEPERATOR);
 	printf ("        corosync-objctl -t object%cchild_obj ...           Track changes\n", SEPERATOR);
 	printf ("        corosync-objctl -a                                Print all objects\n");
+	printf ("        corosync-objctl -p <filename> Load in config from the specified file.\n");
 	printf("\n");
 	return 0;
 }
@@ -203,43 +322,52 @@ static cs_error_t validate_name(char * obj_name_pt)
 		return CS_ERR_INVALID_PARAM;
 }
 
-void get_child_name(char * name_pt, char * child_name)
-{
-	char * tmp;
-	char str_copy[OBJ_NAME_SIZE];
-
-	strcpy(str_copy, name_pt);
-
-	/* first remove the value (it could be a file path */
-	tmp = strchr(str_copy, '=');
-	if (tmp != NULL) *tmp = '\0';
-
-	/* truncate the  */
-	tmp = strrchr(str_copy, SEPERATOR);
-	if (tmp == NULL) tmp = str_copy;
-	strcpy(child_name, tmp+1);
-}
-
-void get_parent_name(char * name_pt, char * parent_name)
+static void get_parent_name(const char * name_pt, char * parent_name)
 {
 	char * tmp;
 	strcpy(parent_name, name_pt);
 
 	/* first remove the value (it could be a file path */
 	tmp = strchr(parent_name, '=');
-	if (tmp != NULL) *tmp = '\0';
+	if (tmp != NULL) {
+		*tmp = '\0';
+		tmp--;
+		while (isblank (*tmp)) {
+			*tmp = '\0';
+			tmp--;
+		}
+	}
 
 	/* then truncate the child name */
 	tmp = strrchr(parent_name, SEPERATOR);
 	if (tmp != NULL) *tmp = '\0';
 }
 
-void get_key(char * name_pt, char * key_name, char * key_value)
+static void get_key(const char * name_pt, char * key_name, char * key_value)
 {
-	char * tmp;
+	char * tmp = (char*)name_pt;
 	char str_copy[OBJ_NAME_SIZE];
+	char * copy_tmp = str_copy;
+	int equals_seen = 0;
+	int in_quotes = 0;
 
-	strcpy(str_copy, name_pt);
+	/* strip out spaces when not in quotes */
+	while (*tmp != '\0') {
+		if (*tmp == '=')
+			equals_seen = 1;
+		if (equals_seen && *tmp == '"') {
+			if (in_quotes)
+				in_quotes = 0;
+			else
+				in_quotes = 1;
+		}
+		if (*tmp != ' ' || in_quotes) {
+			*copy_tmp = *tmp;
+			copy_tmp++;
+		}
+		tmp++;
+	}
+	*copy_tmp = '\0';
 
 	/* first remove the value (it could have a SEPERATOR in it */
 	tmp = strchr(str_copy, '=');
@@ -265,7 +393,7 @@ static cs_error_t find_object (confdb_handle_t handle,
 	hdb_handle_t obj_handle;
 	confdb_handle_t parent_object_handle = OBJECT_PARENT_HANDLE;
 	char tmp_name[OBJ_NAME_SIZE];
-	cs_error_t res;
+	cs_error_t res = CS_OK;
 
 	strncpy (tmp_name, name_pt, OBJ_NAME_SIZE);
 	obj_name_pt = strtok_r(tmp_name, SEPERATOR_STR, &save_pt);
@@ -311,13 +439,18 @@ static void write_key(confdb_handle_t handle, char * path_pt)
 	char key_name[OBJ_NAME_SIZE];
 	char key_value[OBJ_NAME_SIZE];
 	char old_key_value[OBJ_NAME_SIZE];
-	int old_key_value_len;
+	size_t old_key_value_len;
 	cs_error_t res;
+	confdb_value_types_t type;
 
 	/* find the parent object */
 	get_parent_name(path_pt, parent_name);
 	get_key(path_pt, key_name, key_value);
 
+	if (debug == 1)
+		printf ("%d: key:\"%s\", value:\"%s\"\n",
+				__LINE__, key_name, key_value);
+
 	if (validate_name(key_name) != CS_OK) {
 		fprintf(stderr, "Incorrect key name, can not have \"=\" or \"%c\"\n", SEPERATOR);
 		exit(EXIT_FAILURE);
@@ -330,12 +463,11 @@ static void write_key(confdb_handle_t handle, char * path_pt)
 	}
 
 	/* get the current key */
-	res = confdb_key_get (handle,
+	res = confdb_key_get_typed (handle,
 						  obj_handle,
 						  key_name,
-						  strlen(key_name),
 						  old_key_value,
-						  &old_key_value_len);
+						  &old_key_value_len, &type);
 
 	if (res == CS_OK) {
 		/* replace the current value */
@@ -352,12 +484,12 @@ static void write_key(confdb_handle_t handle, char * path_pt)
 			fprintf(stderr, "Failed to replace the key %s=%s. Error %d\n", key_name, key_value, res);
 	} else {
 		/* not there, create a new key */
-		res = confdb_key_create (handle,
+		res = confdb_key_create_typed (handle,
 								 obj_handle,
 								 key_name,
-								 strlen(key_name),
 								 key_value,
-								 strlen(key_value));
+								 strlen(key_value),
+								 CONFDB_VALUETYPE_STRING);
 		if (res != CS_OK)
 			fprintf(stderr, "Failed to create the key %s=%s. Error %d\n", key_name, key_value, res);
 	}
@@ -392,6 +524,9 @@ static void create_object(confdb_handle_t handle, char * name_pt)
 						obj_name_pt);
 				exit(EXIT_FAILURE);
 			}
+
+			if (debug)
+				printf ("%s:%d: %s\n", __func__,__LINE__, obj_name_pt);
 			res = confdb_object_create (handle,
 										parent_object_handle,
 										obj_name_pt,
@@ -407,45 +542,60 @@ static void create_object(confdb_handle_t handle, char * name_pt)
 	}
 }
 
+/* Print "?" in place of any non-printable byte of OBJ. */
+static void print_name (FILE *fp, const void *obj, size_t obj_len)
+{
+	const char *p = obj;
+	size_t i;
+	for (i = 0; i < obj_len; i++) {
+		int c = *p++;
+		if (!isprint (c)) {
+			c = '?';
+		}
+		fputc (c, fp);
+	}
+}
+
 static void tail_key_changed(confdb_handle_t handle,
 	confdb_change_type_t change_type,
 	hdb_handle_t parent_object_handle,
 	hdb_handle_t object_handle,
-	void *object_name_pt,
-	int  object_name_len,
-	void *key_name_pt,
-	int key_name_len,
-	void *key_value_pt,
-	int key_value_len)
+	const void *object_name_pt,
+	size_t  object_name_len,
+	const void *key_name_pt,
+	size_t key_name_len,
+	const void *key_value_pt,
+	size_t key_value_len)
 {
-	char * on = (char*)object_name_pt;
-	char * kn = (char*)key_name_pt;
-	char * kv = (char*)key_value_pt;
-
-	on[object_name_len] = '\0';
-	kv[key_value_len] = '\0';
-	kn[key_name_len] = '\0';
-	printf("key_changed> %s.%s=%s\n", on, kn, kv);
+	/* printf("key_changed> %.*s.%.*s=%.*s\n", */
+	fputs("key_changed> ", stdout);
+	print_name (stdout, object_name_pt, object_name_len);
+	fputs(".", stdout);
+	print_name (stdout, key_name_pt, key_name_len);
+	fputs("=", stdout);
+	print_name (stdout, key_value_pt, key_value_len);
+	fputs("\n", stdout);
 }
 
 static void tail_object_created(confdb_handle_t handle,
 	hdb_handle_t parent_object_handle,
 	hdb_handle_t object_handle,
-	uint8_t *name_pt,
-	int  name_len)
+	const void *name_pt,
+	size_t name_len)
 {
-	name_pt[name_len] = '\0';
-	printf("object_created> %s\n", name_pt);
+	fputs("object_created>", stdout);
+	print_name(stdout, name_pt, name_len);
+	fputs("\n", stdout);
 }
 
 static void tail_object_deleted(confdb_handle_t handle,
 	hdb_handle_t parent_object_handle,
-	uint8_t *name_pt,
-	int  name_len)
+	const void *name_pt,
+	size_t name_len)
 {
-	name_pt[name_len] = '\0';
-
-	printf("object_deleted> %s\n", name_pt);
+	fputs("object_deleted>", stdout);
+	print_name(stdout, name_pt, name_len);
+	fputs("\n", stdout);
 }
 
 static void listen_for_object_changes(confdb_handle_t handle)
@@ -566,11 +716,14 @@ int main (int argc, char *argv[]) {
 	action = ACTION_READ;
 
 	for (;;){
-		c = getopt (argc,argv,"hawcdtp:");
+		c = getopt (argc,argv,"hawcvdtp:");
 		if (c==-1) {
 			break;
 		}
 		switch (c) {
+			case 'v':
+				debug++;
+				break;
 			case 'h':
 				return print_help();
 				break;
@@ -578,9 +731,7 @@ int main (int argc, char *argv[]) {
 				action = ACTION_PRINT_ALL;
 				break;
 			case 'p':
-				printf("%s:%d NOT Implemented yet.\n", __FUNCTION__, __LINE__);
-				return -1;
-				//return read_in_config_file();
+				return read_in_config_file (optarg);
 				break;
 			case 'c':
 				action = ACTION_CREATE;
@@ -648,4 +799,3 @@ int main (int argc, char *argv[]) {
 
 	return 0;
 }
-
